import pytest

from aiobtclientrpc import _errors

rpc_exception_map = {
    r'^The environment is perfectly safe$': ValueError(r'RUN FOR YOUR LIVES!'),
    r'^The (\w+) fell (\w+)$': (ValueError, r'\1: I fell \2!'),
    r'^A (?P<what>\w+) hit the (?P<who>\w+)$': (ValueError, r'\g<who>: I was hit by a \g<what>!'),
    r'\{([^\}]*?)\}': (ValueError, r'<\1>'),
}

@pytest.mark.parametrize(
    argnames='rpc_error, exp_return_value',
    argvalues=(
        (_errors.RPCError('The environment is perfectly safe'), ValueError(r'RUN FOR YOUR LIVES!')),
        (_errors.RPCError('The environment is toppled over'), _errors.RPCError('The environment is toppled over')),

        (_errors.RPCError('The front fell off'), ValueError(r'front: I fell off!')),
        (_errors.RPCError('The bottom fell off'), ValueError(r'bottom: I fell off!')),
        (_errors.RPCError('The front escalated quickly'), _errors.RPCError('The front escalated quickly')),

        (_errors.RPCError('A wave hit the ship'), ValueError(r'ship: I was hit by a wave!')),
        (_errors.RPCError('A whale hit the blimp'), ValueError(r'blimp: I was hit by a whale!')),
        (_errors.RPCError('A whale punched the blimp'), _errors.RPCError('A whale punched the blimp')),

        (_errors.RPCError('Some {text with} braces'), ValueError(r'<text with>')),
        (_errors.RPCError('Some {text} with {braces}'), ValueError(r'<text>')),
    ),
    ids=lambda v: repr(v),
)
def test_RPCError_translate_finds_matching_exception_with_backreferences(rpc_error, exp_return_value):
    return_value = rpc_error.translate(rpc_exception_map)
    assert type(return_value) is type(exp_return_value)
    assert str(return_value) == str(exp_return_value)



import importlib
import os
import shlex
import subprocess
import time

import pytest

import aiobtclientrpc

PACKAGE_NAME = 'tests.' + os.path.basename(os.path.dirname(__file__))


def _get_server_modules():
    server_modules = []
    servers_path = os.path.join(os.path.dirname(__file__), 'servers')
    print(servers_path)

    for filepath in sorted(os.listdir(servers_path)):
        filename = os.path.basename(filepath)
        if not filename.startswith('_') and filename.endswith('.py'):
            module_name = filename[:-len('.py')]
            module = importlib.import_module(f'.{module_name}', package=f'{PACKAGE_NAME}.servers')
            server_modules.append(module)
    return server_modules


@pytest.fixture(
    params=_get_server_modules(), ids=lambda module: module.name,
)
def api(request, tmp_path):
    module = request.param
    info = module.run(tmp_path)

    try:
        run_server(info, tmp_path)
        client = aiobtclientrpc.client(info['client_name'], url=info['server_url'])
        api_module = importlib.import_module('.' + info['client_name'], package=f'{PACKAGE_NAME}.apis')
        api = api_module.API(client)
        yield api
    finally:
        kill_server(info['server_name'])


_running_servers = {}

def run_server(info, tmp_path):
    try:
        info['proc'] = subprocess.Popen(
            shlex.split(info['server_start_cmd']),
            shell=False,
        )
    except FileNotFoundError:
        pytest.skip(info['server_name'] + ' is not installed')

    _running_servers[info['server_name']] = info

    # Check if server_start_cmd failed, e.g. due to invalid arguments
    if info['proc'].poll() is None:
        print('Started', info['server_name'], 'successfully')
        # Wait for RPC interface to come up
        time.sleep(3)

    # Process should still be running (poll() returns None); otherwise report
    # stdout/stderr
    if info['proc'].poll() is not None:
        raise RuntimeError(f'Failed to run {info["server_start_cmd"]}')


def kill_server(name):
    info = _running_servers.get(name, {})
    if info.get('proc', None):
        server_stop_cmd = info.get('server_stop_cmd')
        if server_stop_cmd:
            subprocess.run(shlex.split(server_stop_cmd), shell=False)
        else:
            info['proc'].terminate()
        info['proc'].wait()



import asyncio
import collections
import re
import sys
import weakref
from unittest.mock import Mock, PropertyMock, call

import pytest

from aiobtclientrpc import _base, _errors, _utils

from .common import AsyncMock


class MockURL(_utils.URL):
    pass


class MockRPC(_base.RPCBase):
    name = 'mockbt'
    label = 'MockBT'
    URL = MockURL

    _connect = AsyncMock()
    _disconnect = AsyncMock()
    _call = AsyncMock()


@pytest.mark.parametrize(
    argnames='timeout, exp_timeout, exp_exception',
    argvalues=(
        (1, 1.0, None),
        (2.5, 2.5, None),
        (0, _base.RPCBase.default_timeout, None),
        (None, _base.RPCBase.default_timeout, None),
        (False, _base.RPCBase.default_timeout, None),
        ('foo', None, _errors.ValueError('Invalid timeout')),
    ),
)
def test_timeout(timeout, exp_timeout, exp_exception, mocker):
    rpc = MockRPC()
    rpc.timeout = 123
    mocker.patch.object(rpc, '_invalidate_http_client')
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            rpc.timeout = timeout
        assert rpc._invalidate_http_client.call_args_list == []
    else:
        rpc.timeout = timeout
        assert rpc.timeout == exp_timeout
        assert rpc._invalidate_http_client.call_args_list == [call()]


@pytest.mark.parametrize('attribute', ('scheme', 'host', 'port', 'path', 'username', 'password'))
def test_url(attribute, mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_invalidate_http_client')
    assert isinstance(rpc.url, MockURL)
    assert rpc._invalidate_http_client.call_args_list == []
    setattr(rpc.url, attribute, '123')
    assert getattr(rpc.url, attribute) == '123'
    assert rpc._invalidate_http_client.call_args_list == [call()]
    rpc.url = 'foo://a:b@bar:456/baz'
    assert rpc.url.with_auth == 'foo://a:b@bar:456/baz'
    assert rpc._invalidate_http_client.call_args_list == [call(), call()]


@pytest.mark.parametrize('attribute', ('scheme', 'host', 'port', 'path', 'username', 'password'))
def test_proxy_url(attribute, mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_invalidate_http_client')
    assert rpc.proxy_url is None
    assert rpc._invalidate_http_client.call_args_list == []

    rpc.proxy_url = 'foo://a:b@bar:456/baz'
    assert rpc.proxy_url.with_auth == 'foo://a:b@bar:456/baz'
    assert rpc._invalidate_http_client.call_args_list == [call()]

    setattr(rpc.proxy_url, attribute, '123')
    assert getattr(rpc.proxy_url, attribute) == '123'
    assert rpc._invalidate_http_client.call_args_list == [call(), call()]

    rpc.proxy_url = None
    assert rpc.proxy_url is None
    assert rpc._invalidate_http_client.call_args_list == [call(), call(), call()]


def test_status():
    rpc = MockRPC()
    assert rpc.status is _utils.ConnectionStatus.disconnected
    rpc._status = _utils.ConnectionStatus.connected
    assert rpc.status is _utils.ConnectionStatus.connected

@pytest.mark.parametrize(
    argnames='status, exp_is_connected',
    argvalues=(
        (_utils.ConnectionStatus.connecting, False),
        (_utils.ConnectionStatus.connected, True),
        (_utils.ConnectionStatus.disconnected, False),
    ),
)
def test_is_connected(status, exp_is_connected):
    rpc = MockRPC()
    rpc._status = status
    assert rpc.is_connected is exp_is_connected


@pytest.mark.parametrize('kwargs', ({}, {'foo': 'bar'}))
@pytest.mark.parametrize('args', ((), (1, 2, 3)))
@pytest.mark.parametrize('name', ('connecting', 'connected', 'disconnected'))
def test_connection_callbacks(name, args, kwargs):
    rpc = MockRPC()

    with pytest.raises(AssertionError):
        getattr(rpc, f'set_{name}_callback')('not callable')
    with pytest.raises(AssertionError):
        getattr(rpc, f'unset_{name}_callback')('not callable')

    # Add callback
    callbacks = (Mock(), Mock(), Mock())
    for cb in callbacks:
        getattr(rpc, f'set_{name}_callback')(cb, *args, **kwargs)

    # Call all available callbacks
    rpc._call_connection_callbacks(name)
    others = [n for n in ('connecting', 'connected', 'disconnected') if n != name]
    for other in others:
        rpc._call_connection_callbacks(other)

    assert callbacks[0].call_args_list == [call(*args, **kwargs)]
    assert callbacks[1].call_args_list == [call(*args, **kwargs)]
    assert callbacks[2].call_args_list == [call(*args, **kwargs)]

    # Remove callback
    getattr(rpc, f'unset_{name}_callback')(callbacks[1])

    # Call all available callbacks
    rpc._call_connection_callbacks(name)
    others = [n for n in ('connecting', 'connected', 'disconnected') if n != name]
    for other in others:
        rpc._call_connection_callbacks(other)

    assert callbacks[0].call_args_list == [call(*args, **kwargs), call(*args, **kwargs)]
    assert callbacks[1].call_args_list == [call(*args, **kwargs)]
    assert callbacks[2].call_args_list == [call(*args, **kwargs), call(*args, **kwargs)]


def test_connection_lock():
    rpc = MockRPC()
    for _ in range(3):
        assert rpc._connection_lock is rpc._connection_lock


@pytest.mark.parametrize(
    argnames='raised_exception, exp_exception',
    argvalues=(
        (None, None),
        (asyncio.TimeoutError('Timeout'), _errors.TimeoutError(f'Timeout after {_base.RPCBase.default_timeout} seconds')),
        (_errors.RPCError('No dice'), _errors.RPCError('No dice')),
        (RuntimeError('Unexpected error'), RuntimeError('Unexpected error')),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_connect(raised_exception, exp_exception, mocker):
    rpc = MockRPC()
    cbs = Mock()
    cbs._call_connection_callbacks = Mock()
    cbs._connect = AsyncMock()
    cbs._disconnect = AsyncMock()
    mocker.patch.object(rpc, '_call_connection_callbacks', cbs._call_connection_callbacks)
    mocker.patch.object(rpc, '_connect', cbs._connect)
    mocker.patch.object(rpc, '_disconnect', cbs._disconnect)

    # Connect multiple times concurrently
    connect_calls = (rpc.connect(), rpc.connect())
    # The last connect() call succeeds
    cbs._connect.side_effect = ([raised_exception] * (len(connect_calls) - 1)) + [None]
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await asyncio.gather(*connect_calls)
    else:
        await asyncio.gather(*connect_calls)

    if exp_exception:
        exp_calls_per_connection_attempt = [
            call._call_connection_callbacks('connecting'),
            call._connect(),
            call._disconnect(),
            call._call_connection_callbacks('disconnected'),
        ]
        assert cbs.mock_calls == (
            exp_calls_per_connection_attempt
            * (len(connect_calls) - 1)
        ) + [
            call._call_connection_callbacks('connecting'),
            call._connect(),
            call._call_connection_callbacks('connected'),
        ]
    else:
        assert cbs.mock_calls == [
            call._call_connection_callbacks('connecting'),
            call._connect(),
            call._call_connection_callbacks('connected'),
        ]
    assert rpc.status is _utils.ConnectionStatus.connected


@pytest.mark.parametrize('status', (
    _utils.ConnectionStatus.connecting,
    _utils.ConnectionStatus.connected,
    _utils.ConnectionStatus.disconnected,
), ids=lambda v: str(v),
)
@pytest.mark.parametrize(
    argnames='raised_exception, exp_exception',
    argvalues=(
        (None, None),
        (asyncio.TimeoutError('Timeout'), _errors.TimeoutError(f'Timeout after {_base.RPCBase.default_timeout} seconds')),
        (_errors.ConnectionError('Connection lost'), _errors.ConnectionError('Connection lost')),
        (_errors.AuthenticationError('Password was changed'), _errors.AuthenticationError('Password was changed')),
        (_errors.RPCError('No dice'), _errors.RPCError('No dice')),
        (RuntimeError('Unexpected error'), RuntimeError('Unexpected error')),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_disconnect(raised_exception, exp_exception, status, mocker):
    rpc = MockRPC()
    rpc._status = status
    cbs = Mock()
    cbs._call_connection_callbacks = Mock()
    cbs._disconnect = AsyncMock()
    cbs._close_http_client = AsyncMock()
    mocker.patch.object(rpc, '_call_connection_callbacks', cbs._call_connection_callbacks)
    mocker.patch.object(rpc, '_disconnect', cbs._disconnect)
    mocker.patch.object(rpc, '_close_http_client', cbs._close_http_client)

    # Disconnect multiple times concurrently
    disconnect_calls = (rpc.disconnect(), rpc.disconnect(), rpc.disconnect(), rpc.disconnect())
    cbs._disconnect.side_effect = raised_exception
    if exp_exception and status is not _utils.ConnectionStatus.disconnected:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await asyncio.gather(*disconnect_calls)
    else:
        await asyncio.gather(*disconnect_calls)

    if status is _utils.ConnectionStatus.disconnected:
        # We were already disconnected, but because we are thorough, we made
        # extra sure the HTTP client was closed.
        assert cbs.mock_calls == [
            call._close_http_client(),
        ] * len(disconnect_calls)

    else:
        # _disconnect() and the callback are only called if we weren't already
        # disconnected. The first rpc.disconnect() did the trick and the others
        # just closed the client (which should do nothing if it's already
        # closed).
        assert cbs.mock_calls == [
            call._disconnect(),
            call._call_connection_callbacks('disconnected'),
            call._close_http_client(),
        ] + ([call._close_http_client()] * (len(disconnect_calls) - 1))

    assert rpc.status is _utils.ConnectionStatus.disconnected


@pytest.mark.parametrize(
    argnames='status, exp_connect_calls, raised_exception, exp_exception',
    argvalues=(
        (_utils.ConnectionStatus.connecting, [call()], None, None),
        (_utils.ConnectionStatus.connected, [], None, None),
        (_utils.ConnectionStatus.disconnected, [call()], None, None),
        (_utils.ConnectionStatus.connected, [],
         asyncio.TimeoutError('Timeout'), _errors.TimeoutError(f'Timeout after {_base.RPCBase.default_timeout} seconds')),
    ),
)
@pytest.mark.asyncio
async def test_call(status, exp_connect_calls, raised_exception, exp_exception, mocker):
    rpc = MockRPC()
    rpc._status = status
    mocker.patch.object(rpc, 'connect', AsyncMock())
    mocker.patch.object(rpc, '_call', AsyncMock(side_effect=raised_exception))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc.call('foo', bar='baz', x=24)
    else:
        return_value = await rpc.call('foo', bar='baz', x=24)
        assert return_value is rpc._call.return_value
    assert rpc.connect.call_args_list == exp_connect_calls
    assert rpc._call.call_args_list == [call('foo', bar='baz', x=24)]


@pytest.mark.asyncio
async def test_add_event_handler_only_accepts_callable_handlers(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_subscribe', AsyncMock())
    with pytest.raises(AssertionError, match='^not callable$'):
        await rpc.add_event_handler('foo', 'not callable')
    assert rpc._subscribe.call_args_list == []
    assert rpc._event_handlers == {}

@pytest.mark.asyncio
async def test_add_event_handler_subscribes_to_event(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_subscribe', AsyncMock())
    await rpc.add_event_handler('foo', Mock())
    assert rpc._subscribe.call_args_list == [call('foo')]
    await rpc.add_event_handler('bar', Mock())
    assert rpc._subscribe.call_args_list == [call('foo'), call('bar')]
    await rpc.add_event_handler('foo', Mock())
    assert rpc._subscribe.call_args_list == [call('foo'), call('bar')]
    await rpc.add_event_handler('bar', Mock())
    assert rpc._subscribe.call_args_list == [call('foo'), call('bar')]
    await rpc.add_event_handler('baz', Mock())
    assert rpc._subscribe.call_args_list == [call('foo'), call('bar'), call('baz')]

@pytest.mark.asyncio
async def test_add_event_handler_associates_handler_with_event(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_subscribe', AsyncMock())
    rpc._event_handlers.clear()
    handler1, handler2 = Mock(), Mock()

    await rpc.add_event_handler('foo', handler1)
    assert rpc._event_handlers == {'foo': [handler1]}
    await rpc.add_event_handler('foo', handler2)
    assert rpc._event_handlers == {'foo': [handler1, handler2]}
    await rpc.add_event_handler('foo', handler1)
    assert rpc._event_handlers == {'foo': [handler1, handler2]}
    await rpc.add_event_handler('bar', handler1)
    assert rpc._event_handlers == {'foo': [handler1, handler2], 'bar': [handler1]}
    await rpc.add_event_handler('bar', handler2)
    assert rpc._event_handlers == {'foo': [handler1, handler2], 'bar': [handler1, handler2]}
    await rpc.add_event_handler('bar', handler2)
    assert rpc._event_handlers == {'foo': [handler1, handler2], 'bar': [handler1, handler2]}

@pytest.mark.asyncio
async def test_add_event_handler_with_autoremove(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_subscribe', AsyncMock())
    rpc._event_handlers.clear()
    handler1, handler2, handler3 = Mock(id='h1'), Mock(id='h2'), Mock(id='h3')

    ref_mock = mocker.patch('weakref.ref', Mock(
        side_effect=lambda handler, callback: f'{handler.id}.ref',
    ))

    await rpc.add_event_handler('foo', handler1, autoremove=True)
    assert rpc._event_handlers == {'foo': ['h1.ref']}
    assert ref_mock.call_args_list == [
        call(handler1, rpc._event_handlers['foo'].remove),
    ]

    await rpc.add_event_handler('foo', handler2, autoremove=False)
    assert rpc._event_handlers == {'foo': ['h1.ref', handler2]}
    assert ref_mock.call_args_list == [
        call(handler1, rpc._event_handlers['foo'].remove),
    ]

    await rpc.add_event_handler('bar', handler3, autoremove=False)
    assert rpc._event_handlers == {'foo': ['h1.ref', handler2], 'bar': [handler3]}
    assert ref_mock.call_args_list == [
        call(handler1, rpc._event_handlers['foo'].remove),
    ]

    # Adding the same handler with autoremove multiple times
    for i in range(1, 4):
        await rpc.add_event_handler('bar', handler2, autoremove=True)
        assert rpc._event_handlers == {'foo': ['h1.ref', handler2], 'bar': [handler3, 'h2.ref']}
        assert ref_mock.call_args_list == [
            call(handler1, rpc._event_handlers['foo'].remove),
        ] + [
            call(handler2, rpc._event_handlers['bar'].remove),
        ] * i


@pytest.mark.asyncio
async def test_remove_event_handler(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_unsubscribe', AsyncMock())
    handler1, handler2, handler_ = Mock(), Mock(), Mock()
    rpc._event_handlers.clear()
    rpc._event_handlers.update({'foo': [handler1, handler2], 'bar': [handler2]})

    await rpc.remove_event_handler('foo', handler_)
    assert rpc._event_handlers == {'foo': [handler1, handler2], 'bar': [handler2]}
    assert rpc._unsubscribe.call_args_list == []

    await rpc.remove_event_handler('bar', handler_)
    assert rpc._event_handlers == {'foo': [handler1, handler2], 'bar': [handler2]}
    assert rpc._unsubscribe.call_args_list == []

    await rpc.remove_event_handler('foo', handler1)
    assert rpc._event_handlers == {'foo': [handler2], 'bar': [handler2]}
    assert rpc._unsubscribe.call_args_list == []

    await rpc.remove_event_handler('foo', handler2)
    assert rpc._event_handlers == {'bar': [handler2]}
    assert rpc._unsubscribe.call_args_list == [call('foo')]

    await rpc.remove_event_handler('bar', handler2)
    assert rpc._event_handlers == {}
    assert rpc._unsubscribe.call_args_list == [call('foo'), call('bar')]


@pytest.mark.parametrize('wait_fails', (True, False))
@pytest.mark.asyncio
async def test_wait_for_event(wait_fails, mocker):
    event_name = 'foo'
    rpc = MockRPC()
    mocker.patch.object(rpc, '_subscribe', AsyncMock())
    mocker.patch.object(rpc, '_unsubscribe', AsyncMock())

    if wait_fails:
        Event_mock = mocker.patch('asyncio.Event', return_value=Mock(
            set=Mock(),
            wait=AsyncMock(side_effect=RuntimeError('Something went wrong')),
        ))

    rpc._event_handlers.clear()

    asyncio.create_task(rpc._emit_event(event_name, (1, 2, 3), {'this': 'that'}))
    if wait_fails:
        with pytest.raises(Exception):
            await rpc.wait_for_event(event_name)
    else:
        await rpc.wait_for_event(event_name)

    if wait_fails:
        assert Event_mock.call_args_list == [call()]
    assert rpc._subscribe.call_args_list == [call(event_name)]
    assert rpc._unsubscribe.call_args_list == [call(event_name)]
    assert rpc._event_handlers == {}


@pytest.mark.asyncio
async def test_emit_event_makes_calls_in_order(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_unsubscribe', AsyncMock())
    mocks = Mock()
    mocks.handler1 = Mock()
    mocks.handler2 = AsyncMock()
    mocks.handler3 = Mock()
    mocks.handler4 = AsyncMock()
    rpc._event_handlers.clear()
    rpc._event_handlers.update({
        'foo': [mocks.handler1, mocks.handler2],
        'bar': [mocks.handler3, mocks.handler4],
    })

    await rpc._emit_event('foo', (1, 2, 3), {'this': 'that'})
    assert mocks.mock_calls == [
        call.handler1(1, 2, 3, this='that'),
        call.handler2(1, 2, 3, this='that'),
    ]
    mocks.reset_mock()
    await rpc._emit_event('bar', (4, 5, 6), {'hey': 'ho'})
    assert mocks.mock_calls == [
        call.handler3(4, 5, 6, hey='ho'),
        call.handler4(4, 5, 6, hey='ho'),
    ]

@pytest.mark.asyncio
async def test_emit_event_handles_weak_references(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, '_unsubscribe', AsyncMock())

    calls = collections.defaultdict(lambda: [])

    def handler1(*args, **kwargs):
        calls['handler1'].append(call(*args, **kwargs))

    async def handler2(*args, **kwargs):
        calls['handler2'].append(call(*args, **kwargs))

    def handler3(*args, **kwargs):
        calls['handler3'].append(call(*args, **kwargs))

    async def handler4(*args, **kwargs):
        calls['handler4'].append(call(*args, **kwargs))

    rpc._event_handlers.clear()
    rpc._event_handlers.update({
        'foo': [handler1, weakref.ref(handler2)],
        'bar': [weakref.ref(handler3), handler4],
    })

    await rpc._emit_event('foo', (1, 2, 3), {'this': 'that'})
    assert dict(calls) == {
        'handler1': [call(1, 2, 3, this='that')],
        'handler2': [call(1, 2, 3, this='that')],
    }
    await rpc._emit_event('bar', (4, 5, 6), {'hey': 'ho'})
    assert dict(calls) == {
        'handler1': [call(1, 2, 3, this='that')],
        'handler2': [call(1, 2, 3, this='that')],
        'handler3': [call(4, 5, 6, hey='ho')],
        'handler4': [call(4, 5, 6, hey='ho')],
    }

    del handler1
    del handler2
    del handler3
    del handler4
    for name in ('handler1', ' handler2', ' handler3', ' handler4'):
        assert name not in locals()
    calls.clear()

    await rpc._emit_event('foo', (1, 2, 3), {'this': 'that'})
    assert dict(calls) == {
        'handler1': [call(1, 2, 3, this='that')],
    }
    await rpc._emit_event('bar', (2, 3, 4), {'yo': 'bro'})
    assert dict(calls) == {
        'handler1': [call(1, 2, 3, this='that')],
        'handler4': [call(2, 3, 4, yo='bro')],
    }

@pytest.mark.parametrize(
    argnames='event_handler, exception, get_running_loop',
    argvalues=(
        (Mock(), ValueError('Wat?'), Mock()),
        (Mock(), ValueError('Wat?'), Mock(side_effect=RuntimeError('No running loop'))),
        (AsyncMock(), ValueError('Wat?'), Mock()),
        (AsyncMock(), ValueError('Wat?'), Mock(side_effect=RuntimeError('No running loop'))),
    ),
)
@pytest.mark.asyncio
async def test_emit_event_catches_exceptions_from_handler(event_handler, exception, get_running_loop, mocker):
    rpc = MockRPC()
    rpc._event_handlers.clear()
    rpc._event_handlers.update({
        'foo': [event_handler],
    })
    event_handler.side_effect = exception
    mocker.patch('asyncio.get_running_loop', get_running_loop)

    with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
        await rpc._emit_event('foo', (1, 2, 3), {'this': 'that'})

    if get_running_loop.side_effect:
        assert get_running_loop.return_value.stop.call_args_list == []
    else:
        assert get_running_loop.return_value.stop.call_args_list == [call()]


def test_event_handlers(mocker):
    rpc = MockRPC()
    assert rpc._event_handlers is rpc._event_handlers


@pytest.mark.asyncio
async def test_subscribe(mocker):
    rpc = MockRPC()
    with pytest.raises(NotImplementedError, match=rf'^Events are not supported for {rpc.label}$'):
        await rpc._subscribe('asdf')


@pytest.mark.asyncio
async def test_unsubscribe(mocker):
    rpc = MockRPC()
    with pytest.raises(NotImplementedError, match=rf'^Events are not supported for {rpc.label}$'):
        await rpc._unsubscribe('asdf')


@pytest.mark.asyncio
async def test_context_manager_behaviour(mocker):
    rpc = MockRPC()
    mocker.patch.object(rpc, 'disconnect', AsyncMock())
    mocker.patch.object(rpc, '_close_http_client', AsyncMock())

    # Context manager is reusable
    for i in range(3):
        assert rpc.disconnect.call_args_list == [call()] * i
        assert rpc._close_http_client.call_args_list == [call()] * i
        async with rpc as target:
            assert target is rpc
        assert rpc.disconnect.call_args_list == [call()] * (i + 1)
        assert rpc._close_http_client.call_args_list == [call()] * (i + 1)


def test_http_headers():
    rpc = MockRPC()
    for i in range(3):
        assert rpc._http_headers is rpc._http_headers

    if sys.version_info >= (3, 11):
        exp_msg = "property '_http_headers' of 'MockRPC' object has no setter"
    elif sys.version_info >= (3, 10):
        exp_msg = "can't set attribute '_http_headers'"
    else:
        exp_msg = "can't set attribute"

    with pytest.raises(AttributeError, match=rf'^{re.escape(exp_msg)}$'):
        rpc._http_headers = 'asdf'


@pytest.mark.parametrize('client_is_invalidated', (None, False, True))
@pytest.mark.parametrize('client', (None, Mock()))
@pytest.mark.parametrize('proxy_url', (None, 'mock proxy url'))
@pytest.mark.asyncio
async def test_get_http_client(client_is_invalidated, client, proxy_url, mocker):
    calls = Mock(_close_http_client=AsyncMock())
    rpc = MockRPC()
    if client_is_invalidated is not None:
        rpc._http_client_is_invalidated = client_is_invalidated
    if client:
        rpc._http_client = client
    mocker.patch.object(rpc, '_close_http_client', calls._close_http_client)
    mocker.patch('aiobtclientrpc._utils.create_http_client', calls.create_http_client)
    mocker.patch.object(type(rpc), 'proxy_url', PropertyMock(return_value=Mock(with_auth=proxy_url)))

    return_value = await rpc._get_http_client()
    if client_is_invalidated:
        if client:
            assert calls.mock_calls == [
                call._close_http_client(),
            ]
            assert return_value is client
        else:
            assert calls.mock_calls == [
                call._close_http_client(),
                call.create_http_client(
                    auth=(rpc.url.username, rpc.url.password),
                    proxy_url=proxy_url,
                ),
            ]
            assert return_value is calls.create_http_client.return_value
    else:
        if client:
            assert calls.mock_calls == []
            assert return_value is client
        else:
            assert calls.mock_calls == [
                call.create_http_client(
                    auth=(rpc.url.username, rpc.url.password),
                    proxy_url=proxy_url,
                ),
            ]
            assert return_value is calls.create_http_client.return_value


@pytest.mark.parametrize('client', (None, Mock(aclose=AsyncMock())))
@pytest.mark.asyncio
async def test_close_http_client(client):
    rpc = MockRPC()
    if client is not None:
        rpc._http_client = client
    await rpc._close_http_client()
    if client is not None:
        assert client.aclose.call_args_list == [call()]
    assert not hasattr(rpc, '_http_client')


@pytest.mark.parametrize(
    argnames='has_client, exp_http_client_is_invalidated',
    argvalues=(
        (False, None),
        (True, True),
    ),
)
def test_invalidate_http_client(has_client, exp_http_client_is_invalidated):
    rpc = MockRPC()
    if has_client:
        rpc._http_client = Mock()
    rpc._invalidate_http_client()
    if not exp_http_client_is_invalidated:
        assert not hasattr(rpc, '_http_client_is_invalidated')
    else:
        assert rpc._http_client_is_invalidated is True

    assert rpc.status is _utils.ConnectionStatus.disconnected


@pytest.mark.parametrize(
    argnames='data, exp_kwargs',
    argvalues=(
        (None, {'data': None, 'content': None}),
        ({'foo': 'bar'}, {'data': {'foo': 'bar'}, 'content': None}),
        (['foo', 'bar'], {'data': None, 'content': ['foo', 'bar']}),
        ('foo bar', {'data': None, 'content': 'foo bar'}),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_send_post_request(data, exp_kwargs, mocker):
    rpc = MockRPC()
    client = Mock(post=Mock())
    mocker.patch.object(rpc, '_get_http_client', AsyncMock(return_value=client))
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions', AsyncMock())
    rpc._http_headers['foo'] = 'bar'

    return_value = await rpc._send_post_request(
        url='mock url',
        data=data,
        files='mock files',
    )

    assert return_value is catch_connection_exceptions_mock.return_value
    assert rpc._get_http_client.call_args_list == [call()]
    assert catch_connection_exceptions_mock.call_args_list == [call(client.post.return_value)]
    assert client.post.call_args_list == [call(
        timeout=float('inf'),
        url='mock url',
        headers={'foo': 'bar'},
        files='mock files',
        **exp_kwargs,
    )]



from . import _base, _errors, _utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class QbittorrentURL(_utils.URL):
    """qBittorrent RPC URL"""

    default = 'http://localhost:8080'

    @property
    def scheme(self):
        """Valid schemes: ``http``, ``https``"""
        return super().scheme

    @scheme.setter
    def scheme(self, scheme):
        if scheme and scheme.lower() not in ('http', 'https'):
            raise _errors.ValueError('Scheme must be "http" or "https"')
        else:
            _utils.URL.scheme.fset(self, scheme)

    @property
    def path(self):
        """Always `None`"""
        return super().path

    @path.setter
    def path(self, path):
        if path:
            raise _errors.ValueError("qBittorrent URLs don't have a path")
        else:
            _utils.URL.path.fset(self, path)


class QbittorrentRPC(_base.RPCBase):
    """
    RPC client for qBittorrent

    URL format: ``[http[s]://][USERNAME:PASSWORD@]HOST[:PORT]``

    Reference: https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)

    **Calling RPC methods**

    Passing arguments as keywords:

    >>> call(
    >>>     "torrents/info",
    >>>     hashes="|".join([
    >>>         "d5a34e9eb4709e265f0f03a1c8ab60890dcb94a9",
    >>>         "4435ef55af79b350e7b85d5b330a7886a61e3bdf",
    >>>     ]),
    >>>     sort="size",
    >>>     limit=20,
    >>> )

    Passing arguments as dictionary:

    >>> call(
    >>>     "torrents/info",
    >>>     {
    >>>         "hashes": "|".join([
    >>>             "d5a34e9eb4709e265f0f03a1c8ab60890dcb94a9",
    >>>             "4435ef55af79b350e7b85d5b330a7886a61e3bdf",
    >>>         ]),
    >>>         "sort": "size",
    >>>         "limit" 20,
    >>>     },
    >>> )

    Passing arguments as both keywords and dictionary (keyword values are
    overloaded by dictionary values):

    >>> call(
    >>>     "torrents/info",
    >>>     options={
    >>>         "hashes": "|".join([
    >>>             "d5a34e9eb4709e265f0f03a1c8ab60890dcb94a9",
    >>>             "4435ef55af79b350e7b85d5b330a7886a61e3bdf",
    >>>         ]),
    >>>         "sort": "size",
    >>>     },
    >>>     limit=20,
    >>> )

    The ``files`` argument is special. It is used to add torrents:

    >>> call(
    >>>     'torrents/add',
    >>>     files=[
    >>>         ('filename', (
    >>>             os.path.abspath('path/to/1.torrent'),
    >>>             open('path/to/1.torrent', 'rb'),
    >>>             'application/x-bittorrent',
    >>>         )),
    >>>         ('filename', (
    >>>             os.path.abspath('path/to/2.torrent'),
    >>>             open('path/to/2.torrent', 'rb'),
    >>>             'application/x-bittorrent',
    >>>         )),
    >>>     ],
    >>>     options={
    >>>         'savepath': 'special/download/path',
    >>>         'paused': 'true',
    >>>     },
    >>> )

    :raise ValueError: if any argument is invalid
    """

    name = 'qbittorrent'
    label = 'qBittorrent'
    URL = QbittorrentURL

    def __init__(
        self,
        url=None,
        *,
        scheme=None,
        host=None,
        port=None,
        username=None,
        password=None,
        timeout=None,
        proxy_url=None,
    ):
        # Set custom or default URL
        self.url = url

        # Update URL
        if scheme:
            self.url.scheme = scheme
        if host:
            self.url.host = host
        if port:
            self.url.port = port
        if username:
            self.url.username = username
        if password:
            self.url.password = password

        self.timeout = timeout
        self.proxy_url = proxy_url

    async def _connect(self):
        response = await self._send_post_request(
            url=f'{self.url}/api/v2/auth/login',
            data={
                'username': self.url.username or '',
                'password': self.url.password or '',
            },
        )

        if response.status_code == 403:
            raise _errors.AuthenticationError('Too many failed authentication attempts')
        elif response.text == 'Fails.':
            raise _errors.AuthenticationError('Authentication failed')
        elif response.text != 'Ok.':
            raise _errors.RPCError(response.text)

    async def _disconnect(self):
        if self.is_connected:
            try:
                await self._send_post_request(f'{self.url}/api/v2/auth/logout')
            except _errors.ConnectionError as e:
                _log.debug('Client unreachable while logging out: %r', e)

    async def _call(self, method, options=None, files=None, **kwargs):
        # Merge dictionary arguments with keyword arguments
        if options:
            kwargs.update(options)

        # POST request expects "data" and "files" arguments
        send_post_request_kwargs = {}
        if kwargs:
            send_post_request_kwargs['data'] = kwargs
        if files:
            send_post_request_kwargs['files'] = files

        response = await self._send_post_request(
            url=f'{self.url}/api/v2/{method}',
            **send_post_request_kwargs,
        )

        if response.status_code == 404:
            raise _errors.RPCError('Unknown RPC method')
        elif response.status_code != 200:
            raise _errors.RPCError(response.text)

        try:
            return response.json()
        except ValueError:
            return response.text



import re
from unittest.mock import Mock, PropertyMock, call

import pytest

from aiobtclientrpc import RPCBase, _errors, _qbittorrent, _utils

from .common import AsyncMock, make_url_parts


@pytest.mark.parametrize(
    argnames='url, exp',
    argvalues=(
        ('',
         {'scheme': 'http', 'host': 'localhost', 'port': '8080', 'path': None, 'username': None, 'password': None}),
        ('myhost',
         {'scheme': 'http', 'host': 'myhost', 'port': '8080', 'path': None, 'username': None, 'password': None}),
        ('myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('foo:bar@myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('https://myhost',
         {'scheme': 'https', 'host': 'myhost', 'port': '8080', 'path': None, 'username': None, 'password': None}),
        ('https://myhost:123',
         {'scheme': 'https', 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('https://foo:bar@myhost',
         {'scheme': 'https', 'host': 'myhost', 'port': '8080', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('file://myhost',
         _errors.ValueError('Scheme must be "http" or "https"')),
        ('myhost/foo',
         _errors.ValueError("qBittorrent URLs don't have a path")),
    ),
    ids=lambda v: str(v),
)
def test_QbittorrentURL(url, exp):
    if isinstance(exp, Exception):
        with pytest.raises(type(exp), match=rf'^{re.escape(str(exp))}$'):
            _qbittorrent.QbittorrentURL(url)
    else:
        url = _qbittorrent.QbittorrentURL(url)
        assert make_url_parts(url) == exp


@pytest.mark.parametrize('url', (None, 'http://a:b@foo:123'))
@pytest.mark.parametrize(
    argnames='kwargs',
    argvalues=(
        {},
        {'scheme': 'https'},
        {'host': 'asdf'},
        {'port': '123'},
        {'username': 'this', 'password': 'that'},
        {'timeout': 123},
        {'proxy_url': 'http://hey:ho@bar:456'},
    ),
    ids=lambda v: str(v),
)
def test_instantiation(kwargs, url):
    if url:
        kwargs['url'] = url
    rpc = _qbittorrent.QbittorrentRPC(**kwargs)

    default_url = _utils.URL(_qbittorrent.QbittorrentURL.default)
    exp_url = _utils.URL(_qbittorrent.QbittorrentURL.default)
    if url:
        custom_url = _utils.URL(url)
    for name in ('scheme', 'host', 'port', 'path', 'username', 'password'):
        if name in kwargs:
            exp_value = kwargs[name]
        elif url:
            exp_value = getattr(custom_url, name)
        else:
            exp_value = getattr(default_url, name)
        actual_value = getattr(rpc.url, name)
        assert actual_value == exp_value
        setattr(exp_url, name, exp_value)

    assert rpc.url == exp_url
    assert rpc.timeout == kwargs.get('timeout', RPCBase.default_timeout)
    assert rpc.proxy_url == (_utils.URL(kwargs['proxy_url']) if 'proxy_url' in kwargs else None)

@pytest.mark.parametrize(
    argnames='kwargs, exp_error',
    argvalues=(
        ({'url': 'bar:baz'}, 'Invalid port'),
        ({'port': (1, 2, 3)}, 'Invalid port'),
        ({'timeout': 'never'}, 'Invalid timeout'),
        ({'proxy_url': 'foo://bar:baz'}, 'Invalid port'),
    ),
    ids=lambda v: str(v),
)
def test_instantiation_with_invalid_argument(kwargs, exp_error):
    with pytest.raises(_errors.ValueError, match=rf'^{re.escape(exp_error)}$'):
        _qbittorrent.QbittorrentRPC(**kwargs)


@pytest.mark.parametrize('username', (None, '', 'a'))
@pytest.mark.parametrize('password', (None, '', 'b'))
@pytest.mark.parametrize(
    argnames='response, exp_exception',
    argvalues=(
        (Mock(status_code=403), _errors.AuthenticationError('Too many failed authentication attempts')),
        (Mock(status_code=200, text='Fails.'), _errors.AuthenticationError('Authentication failed')),
        (Mock(status_code=200, text='The Reason.'), _errors.RPCError('The Reason.')),
        (Mock(status_code=123, text='The Reason.'), _errors.RPCError('The Reason.')),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_connect(response, exp_exception, username, password, mocker):
    rpc = _qbittorrent.QbittorrentRPC()
    rpc.url = f'http://{username if username else ""}:{password if password else ""}@foo:123'

    mocker.patch.object(rpc, '_send_post_request', AsyncMock(return_value=response))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._connect()
    else:
        # Raises no exception
        await rpc._connect()

    assert rpc._send_post_request.call_args_list == [call(
        url='http://foo:123/api/v2/auth/login',
        data={
            'username': username if username else '',
            'password': password if password else '',
        },
    )]

@pytest.mark.parametrize('is_connected', (True, False))
@pytest.mark.parametrize('exception', (
    None,
    _errors.ConnectionError('Ignored'),
    _errors.TimeoutError('Raised'),
    _errors.AuthenticationError('Raised'),
    RuntimeError('Raised'),
))
@pytest.mark.asyncio
async def test_disconnect(exception, is_connected, mocker):
    rpc = _qbittorrent.QbittorrentRPC()
    rpc.url = 'http://a:b@foo:123'

    mocker.patch.object(type(rpc), 'is_connected', PropertyMock(return_value=is_connected))
    mocker.patch.object(rpc, '_send_post_request', AsyncMock(side_effect=exception))
    if is_connected and exception and not isinstance(exception, _errors.ConnectionError):
        with pytest.raises(type(exception), match=rf'^{re.escape(str(exception))}$'):
            await rpc._disconnect()
    else:
        await rpc._disconnect()

    if is_connected:
        assert rpc._send_post_request.call_args_list == [call(
            'http://foo:123/api/v2/auth/logout',
        )]
    else:
        assert rpc._send_post_request.call_args_list == []


@pytest.mark.parametrize(
    argnames='options, files, kwargs, exp_send_post_request_kwargs',
    argvalues=(
        (None, None, {}, {}),
        ({'dict': 'data'}, {}, {}, {'data': {'dict': 'data'}}),
        ({}, {'the': 'files'}, {}, {'files': {'the': 'files'}}),
        ({}, {}, {'kw': 'data'}, {'data': {'kw': 'data'}}),
        ({'dict': 'data'}, {'the': 'files'}, {},
         {'data': {'dict': 'data'}, 'files': {'the': 'files'}}),
        ({}, {'the': 'files'}, {'kw': 'data'},
         {'data': {'kw': 'data'}, 'files': {'the': 'files'}}),
        ({'dict': 'data', 'more': 'DATA'}, {}, {'kw': 'data', 'more': 'data'},
         {'data': {'kw': 'data', 'more': 'DATA', 'dict': 'data'}}),
        ({'dict': 'data', 'more': 'DATA'}, {'the': 'files'}, {'kw': 'data', 'more': 'data'},
         {'data': {'kw': 'data', 'more': 'DATA', 'dict': 'data'}, 'files': {'the': 'files'}}),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_call_merges_arguments(options, files, kwargs, exp_send_post_request_kwargs, mocker):
    rpc = _qbittorrent.QbittorrentRPC()
    rpc.url = 'http://a:b@foo:123'
    method = 'do_this'

    mocker.patch.object(rpc, '_send_post_request', AsyncMock(
        return_value=Mock(status_code=200, json=Mock(return_value='mock json data')),
    ))

    await rpc._call(method, options=options, files=files, **kwargs)

    assert rpc._send_post_request.call_args_list == [call(
        url=f'http://foo:123/api/v2/{method}',
        **exp_send_post_request_kwargs,
    )]


@pytest.mark.parametrize(
    argnames='response, exp_exception, exp_return_value',
    argvalues=(
        (Mock(status_code=404), _errors.RPCError('Unknown RPC method'), None),
        (Mock(status_code=123, text='The Error.'), _errors.RPCError('The Error.'), None),
        (Mock(status_code=200, text='The Text.', json=Mock(side_effect=ValueError())), None, 'The Text.'),
        (Mock(status_code=200, json=Mock(return_value='The JSON.')), None, 'The JSON.'),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_call_handles_exceptions(response, exp_exception, exp_return_value, mocker):
    rpc = _qbittorrent.QbittorrentRPC()
    rpc.url = 'http://a:b@foo:123'
    method = 'do_this'

    mocker.patch.object(rpc, '_send_post_request', AsyncMock(return_value=response))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._call(method)
    else:
        return_value = await rpc._call(method)
        assert return_value is exp_return_value

    assert rpc._send_post_request.call_args_list == [call(
        url=f'http://foo:123/api/v2/{method}',
    )]



"""
https://github.com/abhinavsingh/proxy.py/blob/develop/examples/https_connect_tunnel.py

:copyright: (c) 2013-present by Abhinav Singh and contributors.
:license: BSD, see LICENSE for more details.
"""

import contextlib

from proxy import Proxy
from proxy.core.base import BaseTcpTunnelHandler
from proxy.http.responses import (PROXY_TUNNEL_ESTABLISHED_RESPONSE_PKT,
                                  PROXY_TUNNEL_UNSUPPORTED_SCHEME)


class _HttpsConnectTunnelHandler(BaseTcpTunnelHandler):
    """HTTP[S] CONNECT tunnel."""

    def handle_data(self, data):
        # Queue for upstream if connection has been established
        if self.upstream and self.upstream._conn is not None:
            self.upstream.queue(data)
            return None

        # Parse client request
        self.request.parse(data)

        # Drop the request if not a CONNECT request
        if not self.request.is_https_tunnel:
            self.work.queue(PROXY_TUNNEL_UNSUPPORTED_SCHEME)
            return True

        # CONNECT requests are short and we need not worry about
        # receiving partial request bodies here.
        assert self.request.is_complete

        # Establish connection with upstream
        try:
            self.connect_upstream()
        except OSError as e:
            # Close client connection
            import logging
            logging.getLogger(__name__).debug('Failed to connect to upstream: %r', e)
            raise
            # return True
        else:
            # Queue tunnel established response to client
            self.work.queue(PROXY_TUNNEL_ESTABLISHED_RESPONSE_PKT)


@contextlib.contextmanager
def http_proxy(port=60100):
    with Proxy(
        work_klass=_HttpsConnectTunnelHandler,
        port=port,
        num_workers=1,
    ):
        yield f'http://localhost:{port}'

# TODO: Add SOCKS4/5 proxies. proxy.py doesn't seem to support them currently.

proxies = (http_proxy,)



import base64
import os


def get_home(name, tmp_path):
    homepath = tmp_path / f'{name}.home'
    homepath.mkdir(parents=True, exist_ok=True)
    return homepath


def get_torrent_filepath(infohash):
    torrents_dirpath = os.path.join(os.path.dirname(__file__), 'torrents')
    return os.path.join(torrents_dirpath, f'{infohash}.torrent')


def read_torrent_file(filepath):
    with open(filepath, 'rb') as f:
        filecontent = f.read()
    return str(base64.b64encode(filecontent), encoding='ascii')



import asyncio
import enum
import inspect
import os
import re

from . import __project_name__, __version__, _errors

import logging  # isort:skip
_log = logging.getLogger(__name__)


def get_aioloop():
    """Return :class:`asyncio.AbstractEventLoop` instance"""
    # https://docs.python.org/3.10/library/asyncio-eventloop.html
    try:
        return asyncio.get_running_loop()
    except RuntimeError:
        # "no running event loop"
        # We need a loop before the application has started. We can't use
        # get_event_loop(), because that is going to be an alias for
        # get_running_loop() in Python >= 3.10. This is what get_event_loop()
        # does internally in Python 3.6.
        return asyncio.get_event_loop_policy().get_event_loop()


def clients():
    """Return list of :class:`~.RPCBase` subclasses"""
    import aiobtclientrpc  # isort:skip
    basecls = aiobtclientrpc.RPCBase
    subclses = set()
    for name, value in inspect.getmembers(aiobtclientrpc):
        if (
            value is not basecls and
            isinstance(value, type) and
            issubclass(value, basecls)
        ):
            subclses.add(value)
    return sorted(subclses, key=lambda cls: cls.name)


def client(name, *args, **kwargs):
    """
    Convenience function to instantiate a :class:`~.RPCBase` subclass

    :param str name: :attr:`~.RPCBase.name` of the client
    :param args: Positional arguments to pass to the :class:`~.RPCBase` subclass
    :param kwargs: Keyword arguments to pass to the :class:`~.RPCBase` subclass

    :raise ValueError: if there is no :class:`~.RPCBase` subclass with a
        matching `name`

    :return: :class:`~.RPCBase` instance
    """
    for cls in clients():
        if cls.name == name:
            return cls(*args, **kwargs)
    raise _errors.ValueError(f'No such client: {name}')


class ConnectionStatus(enum.Enum):
    """Current state of the client connection"""

    connecting = 'connecting'
    """Attempting to connect"""

    connected = 'connected'
    """Connection was established"""

    disconnected = 'disconnected'
    """Connection was either lost or terminated"""


def cached_property(fget):
    """Property that is replaces itself with its value on first access"""
    class _cached_property():
        def __init__(self, fget):
            self._fget = fget
            self._property_name = fget.__name__

        def __get__(self, obj, cls):
            value = self._fget(obj)
            setattr(obj, self._property_name, value)
            return value

    return _cached_property(fget)


class _URLMeta(type):
    def __init__(cls, name, bases, attrs):
        # Convert default URL to instance of `cls`
        cls.default = cls(cls.default)

class URL(metaclass=_URLMeta):
    """
    URL of an RPC interface

    This implementation attempts to parse URLs more intuitively instead of
    following any specs. For example ``"localhost:1234"`` is interpreted as
    ``host=localhost, port=1234`` instead of ``scheme=localhost, path=1234``.

    :param str url: URL string
    :param str default: Fallback URL when `url` is falsy
    :param callable on_change: Callback that is called with no arguments when
        any property is modified

    :raise ValueError: if `url` is invalid
    """

    _parts = (
        'scheme',
        'username',
        'password',
        'host',
        'port',
        'path',
    )

    @classmethod
    def _dict_from_string(cls, string):
        string = str(string).strip()
        parts = {k: None for k in cls._parts}

        # Scheme
        scheme_regex = re.compile(r'^(.*?)://')
        match = scheme_regex.search(string)
        if match:
            parts['scheme'] = match.group(1) or None
            string = scheme_regex.sub('', string)
        elif string.startswith(os.sep):
            # Assume file system path if URL starts with path separator
            parts['scheme'] = 'file'

        # File system paths don't have authentication, port, etc
        if parts['scheme'] == 'file':
            parts['path'] = string or None

        else:
            # Authentication
            auth_regex = re.compile(r'^(.*?):(.*?)@')
            match = auth_regex.search(string)
            if match:
                parts['username'] = match.group(1) or None
                parts['password'] = match.group(2) or None
                string = auth_regex.sub('', string)

            # Host
            host_regex = re.compile(r'^(.*?)(?=/|:|$)')
            match = host_regex.search(string)
            if match:
                parts['host'] = match.group(1) or None
                string = host_regex.sub('', string)

            # Port
            port_regex = re.compile(r'^:(.*?)(?=/|$)')
            match = port_regex.search(string)
            if match:
                parts['port'] = match.group(1) or None
                string = port_regex.sub('', string)

            # Path
            parts['path'] = string or None

        return parts

    @classmethod
    def _dict_from_any(cls, thing):
        if thing is None:
            return cls._dict_from_string('')

        elif isinstance(thing, str):
            return cls._dict_from_string(thing)

        # URL class is not in the global namespace yet, so we can't do
        # `isinstance(thing, URL)`. Instead, we do duck-typing and rely on
        # `with_auth` method providing what we expect.
        elif hasattr(thing, 'with_auth'):
            return cls._dict_from_string(thing.with_auth)

        else:
            raise TypeError(f'Unsupported type: {type(thing).__name__}: {thing!r}')

    default = ''
    """URL to use when ``url`` and ``default`` arguments are both falsy"""

    def __new__(cls, url=None, default=None, on_change=None):
        # Get instance
        self = super().__new__(cls)

        # Don't trigger any changes until we finished initialization
        self._on_change = None

        default_url = cls._dict_from_any(default or cls.default)
        if url:
            custom_url = cls._dict_from_any(url)
        else:
            custom_url = default_url

        for name in ('scheme', 'username', 'password', 'host', 'port', 'path'):
            if custom_url[name]:
                setattr(self, name, custom_url[name])
            else:
                try:
                    setattr(self, name, default_url[name])
                except _errors.ValueError:
                    setattr(self, name, None)

        # Initialization is complete - enable on_change callback
        self._on_change = on_change

        return self

    @property
    def scheme(self):
        """Scheme (e.g. ``"http"`` or ``"file"``)"""
        return self._scheme

    @scheme.setter
    def scheme(self, scheme):
        if not scheme:
            self._scheme = None
        else:
            self._scheme = str(scheme).lower()

        if self._on_change:
            self._on_change()

    @property
    def host(self):
        """Host name or IP address or `None`"""
        return self._host

    @host.setter
    def host(self, host):
        if not host:
            self._host = None
        else:
            self._host = str(host)

        if self._on_change:
            self._on_change()

    @property
    def port(self):
        """Port number or `None`"""
        return self._port

    @port.setter
    def port(self, port):
        if not port:
            self._port = None
        else:
            try:
                port = int(port)
            except (ValueError, TypeError):
                raise _errors.ValueError('Invalid port')
            else:
                if not 1 <= port <= 65535:
                    raise _errors.ValueError('Invalid port')
                else:
                    self._port = str(port)

        if self._on_change:
            self._on_change()

    @property
    def path(self):
        """File system path or request path or `None`"""
        return self._path

    @path.setter
    def path(self, path):
        self._path = str(path) if path else None

        if self._on_change:
            self._on_change()

    @property
    def username(self):
        """Username for authentication"""
        return self._username

    @username.setter
    def username(self, username):
        if not username:
            self._username = None
        else:
            self._username = str(username)

        if self._on_change:
            self._on_change()

    @property
    def password(self):
        """Password for authentication"""
        return self._password

    @password.setter
    def password(self, password):
        if not password:
            self._password = None
        else:
            self._password = str(password)

        if self._on_change:
            self._on_change()

    @property
    def without_auth(self):
        """URL string without :attr:`username` and :attr:`password`"""
        return self._as_string(with_auth=False)

    @property
    def with_auth(self):
        """URL string with :attr:`username` and :attr:`password`"""
        return self._as_string(with_auth=True)

    def _as_string(self, with_auth=False):
        parts = []

        if self.scheme:
            parts.append(f'{self.scheme}://')

        if with_auth:
            if self.username or self.password:
                parts.append(f'{self.username or ""}:{self.password or ""}@')

        if self.host:
            parts.append(self.host)

        if self.port:
            parts.append(f':{self.port}')

        if self.path:
            parts.append(self.path)

        return ''.join(parts)

    def __str__(self):
        return self.without_auth

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.with_auth == other.with_auth
        else:
            return NotImplemented

    def __ne__(self, other):
        eq = self.__eq__(other)
        return NotImplemented if eq is NotImplemented else not eq

    def __repr__(self):
        text = f'{type(self).__name__}({str(self)!r}'
        if self._on_change:
            text += f', on_change={self._on_change!r}'
        text += ')'
        return text


def create_http_client(*, auth=(None, None), proxy_url=None):
    """
    Return :class:`httpx.AsyncClient` instance

    :param auth: Basic auth credentials as `(username, password)` tuple; if
        either value is falsy, don't do authentication
    :param proxy_url: URL of a SOCKS4, SOCKS5 or HTTP proxy
    """
    import httpx  # isort:skip

    kwargs = {
        # Timeouts are handled with async_timeout in RPCBase
        'timeout': float('inf'),
        'headers': {
            'User-Agent': f'{__project_name__} {__version__}',
        },
    }

    # Basic auth
    username, password = auth
    if username and password:
        kwargs['auth'] = httpx.BasicAuth(username, password)

    # SOCKS[4|5] or HTTP proxy
    if proxy_url:
        import httpx_socks  # isort:skip
        kwargs['transport'] = httpx_socks.AsyncProxyTransport.from_url(proxy_url)

    return httpx.AsyncClient(**kwargs)


async def catch_connection_exceptions(coro):
    """
    Turn exceptions from network requests into :class:`~.ConnectionError`

    Proxy exceptions are also caught.

    The error message should be user-friendly.

    :param coro: Awaitable that performs a network request

    :return: return value of `coro`

    :raise ConnectionError: if any relevant exception is raised
    """
    import httpx, httpx_socks  # isort:skip

    def prettify_msg(e):
        msg = e.strerror if getattr(e, 'strerror', None) else str(e)
        if not msg:
            msg = 'Unknown error'
        else:
            # Extract actual error, e.g. from
            # [Errno 111] Connect call failed ('::1', 5001, 0, 0)
            # Multiple exceptions: [Errno 111] Connect call failed ('::1', 5001, 0, 0),
            #                      [Errno 111] Connect call failed ('127.0.0.1', 5001)
            match = re.search(r'\[Errno -?\d+\]\s*(.*?)\s*(?:\[|\(|$)', msg)
            if match:
                msg = match.group(1)
        return msg

    try:
        return await coro
    except httpx.HTTPError as e:
        raise _errors.ConnectionError(prettify_msg(e))
    except httpx_socks.ProxyError as e:
        raise _errors.ConnectionError(prettify_msg(e))
    except ConnectionAbortedError:
        raise _errors.ConnectionError('Connection aborted')
    except ConnectionRefusedError:
        raise _errors.ConnectionError('Connection refused')
    except ConnectionResetError:
        raise _errors.ConnectionError('Connection reset')
    except OSError as e:
        # Any low-level exceptions and httpx_socks.ProxyConnectionError, which
        # is a subclass of OSError.
        raise _errors.ConnectionError(prettify_msg(e))


_basic_types = {
    bool,
    bytearray,
    bytes,
    dict,
    float,
    int,
    list,
    str,
    tuple,
    type(None),
}

def convert_to_basic_type(data):
    """
    Convert subclassed instance of basic types (:class:`str`, :class:`list`,
    etc) to parent type

    Mappings and sequences are converted recursively.
    """
    data_type = type(data)
    if data_type == list:
        return [
            convert_to_basic_type(item)
            for item in data
        ]
    elif data_type == tuple:
        return tuple(
            convert_to_basic_type(item)
            for item in data
        )
    elif data_type == dict:
        return {
            convert_to_basic_type(k) : convert_to_basic_type(v)
            for k, v in data.items()
        }
    elif data_type in _basic_types:
        return data

    else:
        for t in _basic_types:
            if isinstance(data, t):
                return convert_to_basic_type(t(data))

        raise TypeError(f'Unsupported basic type: {type(data).__name__}: {data!r}')



import asyncio
import sys
import time
from unittest.mock import Mock, call, patch

import pytest

import aiobtclientrpc

from . import common, proxyserver

import logging  # isort:skip
_log = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_authentication_error(api, tmp_path):
    if not api.client.url.username:
        pytest.skip(f'No authentication: {api.client.url}')

    else:
        async with api.client:
            await api.perform_simple_request()

        correct_username = api.client.url.username
        api.client.url.username = 'wrong_username'
        async with api.client:
            with pytest.raises(aiobtclientrpc.AuthenticationError, match=r'^Authentication failed$'):
                await api.perform_simple_request()
        api.client.url.username = correct_username

        correct_password = api.client.url.password
        api.client.url.password = 'wrong_password'
        async with api.client:
            with pytest.raises(aiobtclientrpc.AuthenticationError, match=r'^Authentication failed$'):
                await api.perform_simple_request()
        api.client.url.password = correct_password

        async with api.client:
            await api.perform_simple_request()


@pytest.mark.asyncio
async def test_api_as_context_manager(api, tmp_path):
    for _ in range(3):
        async with api.client:
            await api.perform_simple_request()


@pytest.mark.parametrize('paused', (True, False), ids=lambda paused: 'paused' if paused else 'started')
@pytest.mark.parametrize('as_file', (True, False), ids=lambda as_file: 'as_file' if as_file else 'as_bytes')
@pytest.mark.asyncio
async def test_add_torrents(as_file, paused, api, tmp_path):
    infohashes = sorted([
        '4435ef55af79b350e7b85d5b330a7886a61e3bdf',
        'd5a34e9eb4709e265f0f03a1c8ab60890dcb94a9',
    ])

    try:
        return_value = await api.add_torrent_files(
            torrent_filepaths=(
                common.get_torrent_filepath(infohashes[0]),
                common.get_torrent_filepath(infohashes[1]),
            ),
            as_file=as_file,
            paused=paused,
        )
        assert return_value == infohashes
        torrent_list = await api.get_torrent_list()
        assert torrent_list == infohashes

    finally:
        await api.client.disconnect()


@pytest.mark.skipif(sys.version_info < (3, 8), reason='Avoid BrokenResourceError on Python 3.7')
@pytest.mark.parametrize(
    argnames='start_proxy',
    argvalues=proxyserver.proxies,
)
@pytest.mark.asyncio
async def test_proxy(start_proxy, api, tmp_path):
    with start_proxy() as proxy_url:
        api.client.proxy_url = proxy_url
        _log.debug('proxy started: %s', api.client.proxy_url)

        # Setting a proxy for a file:// URL shouldn't raise immediately (maybe
        # the URL is changed right after), but it should raise when a request is
        # made with an invalid protocol/proxy combination.
        if api.client.url.scheme == 'file':
            with pytest.raises(aiobtclientrpc.ValueError, match=rf'^You cannot use a proxy to connect to {api.client.url}$'):
                await api.perform_simple_request()
            return
        else:
            await api.perform_simple_request()

        _log.debug('stopping proxy: %s', api.client.proxy_url)
    _log.debug('proxy stopped: %s', api.client.proxy_url)

    # Proxy is now stopped while client still thinks it's connected
    if api.client.name == 'deluge':
        exp_error = r'^Connection lost$'
    else:
        exp_error = rf'^Could not connect to proxy {api.client.proxy_url.host}:{api.client.proxy_url.port}$'
    with pytest.raises(aiobtclientrpc.ConnectionError, match=exp_error):
        await api.perform_simple_request()

    # Start proxy again and client should use it
    with start_proxy() as proxy_url:
        _log.debug('proxy started again: %s', proxy_url)
        await api.perform_simple_request()
        _log.debug('stopping proxy: %s', api.client.proxy_url)
    _log.debug('proxy stopped: %s', api.client.proxy_url)

    _log.debug('disconnecting client')
    try:
        await api.client.disconnect()
    except aiobtclientrpc.ConnectionError:
        # qbittorrent's API has a logout() method, so we expect a
        # ConnectionError due to the proxy being down. The other clients don't
        # have a logout() method.
        if api.client.name == 'qbittorrent':
            pass
        else:
            raise

    _log.debug('client is now disconnected')


@pytest.mark.asyncio
async def test_timeout(api, tmp_path):
    # Call the original client._call() method after a nap
    def make_delayed_call(delay, real_call_method=api.client._call):
        async def delayed_call(*args, **kwargs):
            print(time.monotonic(), 'sleeping', delay, 'seconds')
            await asyncio.sleep(delay)
            print(time.monotonic(), 'calling', real_call_method)
            return await real_call_method(*args, **kwargs)

        return delayed_call

    async def connect():
        # Connect without delay
        if not api.client.is_connected:
            print(time.monotonic(), 'connecting')
            await api.client.connect()
            print(time.monotonic(), 'connected')

    # We must connect now because we don't want any autoconnecting behaviour
    # with artifically delayed responses as connect() performs an unpredictable
    # amount of requests. We must set the timeout before connecting manually
    # because it invalidates an existing connection.
    api.client.timeout = 1
    await connect()

    try:
        with patch.object(api.client, '_call', make_delayed_call(0.5)):
            await api.perform_simple_request()

        with patch.object(api.client, '_call', make_delayed_call(100)):
            with pytest.raises(aiobtclientrpc.TimeoutError, match=rf'^Timeout after {api.client.timeout} seconds$'):
                await api.perform_simple_request()
    finally:
        await api.client.disconnect()


@pytest.mark.asyncio
async def test_event_subscriptions_survive_reconnecting(api, tmp_path):
    infohashes = sorted([
        '4435ef55af79b350e7b85d5b330a7886a61e3bdf',
        'd5a34e9eb4709e265f0f03a1c8ab60890dcb94a9',
    ])

    torrent_added_handler = Mock()

    async with api.client:
        try:
            await api.on_torrent_added(torrent_added_handler)
        except NotImplementedError as e:
            assert str(e) == f'Events are not supported for {api.client.label}'
            pytest.skip(str(e))
        else:
            await api.add_torrent_files(
                torrent_filepaths=[common.get_torrent_filepath(infohashes[0])],
            )
            assert torrent_added_handler.call_args_list == [
                call(infohashes[0]),
            ]

    # Re-connect and check if torrent_added_handler() is still called
    async with api.client:
        await api.add_torrent_files(
            torrent_filepaths=[common.get_torrent_filepath(infohashes[1])],
        )
        assert torrent_added_handler.call_args_list == [
            call(infohashes[0]),
            call(infohashes[1]),
        ]


@pytest.mark.asyncio
async def test_waiting_for_event(api, tmp_path):
    infohashes = sorted([
        '4435ef55af79b350e7b85d5b330a7886a61e3bdf',
        'd5a34e9eb4709e265f0f03a1c8ab60890dcb94a9',
    ])

    async with api.client:
        coros = [
            api.wait_for_torrent_added(),
            api.add_torrent_files(
                torrent_filepaths=[
                    common.get_torrent_filepath(infohash)
                    for infohash in infohashes
                ],
            ),
        ]

        return_values = await asyncio.gather(*coros, return_exceptions=True)
        assert len(return_values) == 2
        wait_for_torrent_added_return_value = return_values[0]
        add_torrent_files_return_value = return_values[1]

        if wait_for_torrent_added_return_value is not None:
            assert type(wait_for_torrent_added_return_value) is NotImplementedError
            assert str(wait_for_torrent_added_return_value) == f'Events are not supported for {api.client.label}'
            pytest.skip(str(wait_for_torrent_added_return_value))

        else:
            assert add_torrent_files_return_value == infohashes



import asyncio
import struct

from . import _base, _errors, _utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class DelugeURL(_utils.URL):
    """Deluge RPC URL"""

    default = 'localhost:58846'

    @property
    def scheme(self):
        """Valid schemes: ``None``"""
        return super().scheme

    @scheme.setter
    def scheme(self, scheme):
        if scheme:
            raise _errors.ValueError("Deluge URLs don't have a scheme")
        else:
            _utils.URL.scheme.fset(self, scheme)

    @property
    def path(self):
        """Always `None`"""
        return super().path

    @path.setter
    def path(self, path):
        if path:
            raise _errors.ValueError("Deluge URLs don't have a path")
        else:
            _utils.URL.path.fset(self, path)


class DelugeRPC(_base.RPCBase):
    """
    RPC client for Deluge

    URL format: ``[USERNAME:PASSWORD@]HOST[:PORT]``

    References:
        https://deluge.readthedocs.io/en/latest/reference/index.html
        https://www.rasterbar.com/products/libtorrent/manual-ref.html
        https://git.deluge-torrent.org/deluge/tree/

    **RPC methods**

    RPC methods are only documented as Deluge code. Look for funtions decorated
    with @export. The RPC method is the module name and the function name
    concatenated with ".".

    For example, the RPC method name of Daemon.get_method_list() in
    deluge/core/daemon.py would be "daemon.get_method_list".

    Arguments for RPC methods must be positional/keyword as specified in the
    function's call signature.

    **Events**

    Like RPC methods, events are not properly documented. You can find event
    names by grepping for the class name ``DelugeEvent``. The names of
    subclasses of ``DelugeEvent`` are also event names.

    .. warning:: The Deluge daemon does not complain about invalid event names
        and silently accepts subscribtions to anything. Check your event names
        carefully!

    :raise ValueError: if any argument is invalid
    """

    name = 'deluge'
    label = 'Deluge'
    URL = DelugeURL

    def __init__(
        self,
        url=None,
        *,
        host=None,
        port=None,
        username=None,
        password=None,
        timeout=None,
        proxy_url=None,
    ):
        # Set custom or default URL
        self.url = url

        # Update URL
        if host:
            self.url.host = host
        if port:
            self.url.port = port
        if username:
            self.url.username = username
        if password:
            self.url.password = password

        self.timeout = timeout
        self.proxy_url = proxy_url

    async def _connect(self):
        # Close old client
        await self._disconnect()

        # Create new client
        self._client = _DelugeRPCClient(
            host=self.url.host,
            port=self.url.port,
            on_connection_lost=self._on_connection_lost,
            proxy_url=self.proxy_url,
            event_handler=self._emit_event,
        )
        await _utils.catch_connection_exceptions(
            self._client.login(
                username=self.url.username,
                password=self.url.password,
            ),
        )

        # Subscribe to events again
        for event_name in self._event_handlers:
            # IMPORTANT: We can't use _subscribe() because it uses
            # RPCBase.call(), which will see we're not connected yet and acquire
            # the connection lock to connect. But the connection lock is already
            # locked until this method returns, resulting in a deadlock.
            await self._call('daemon.set_event_interest', [event_name])

    def _on_connection_lost(self):
        self._status = _utils.ConnectionStatus.disconnected
        self._call_connection_callbacks('disconnected')

    async def _disconnect(self):
        if hasattr(self, '_client'):
            await _utils.catch_connection_exceptions(
                self._client.logout()
            )
            delattr(self, '_client')

    async def _call(self, method, *args, **kwargs):
        return await _utils.catch_connection_exceptions(
            self._client.call(method, *args, **kwargs),
        )

    async def _subscribe(self, event_name):
        _log.debug('Setting interest for event: %r', event_name)
        await self.call('daemon.set_event_interest', [event_name])

    async def _unsubscribe(self, event_name):
        # This is not supported in Deluge, but we can just ignore the event by
        # not having any callbacks. This is handled in RPCBase.
        pass


class _DelugeRPCClient:
    def __init__(self, host, port, proxy_url=None, on_connection_lost=None, event_handler=None):
        self._host = str(host)
        self._port = int(port)
        self._on_connection_lost = on_connection_lost
        self._proxy_url = proxy_url

        self._loop = _utils.get_aioloop()
        self._protocol = None
        self._event_handler = event_handler

    def _connection_lost(self):
        self._protocol = None
        if self._on_connection_lost:
            self._on_connection_lost()

    def _event_received(self, event_name, args):
        if asyncio.iscoroutinefunction(self._event_handler):
            coro = self._event_handler(event_name, args)
            self._loop.create_task(coro)
        else:
            self._loop.call_soon(self._event_handler, event_name, args)

    def _protocol_factory(self):
        return _DelugeRPCProtocol(
            on_connection_lost=self._connection_lost,
            on_event_received=self._event_received,
        )

    def _create_ssl_context(self):
        import ssl  # isort:skip
        ctx = ssl.create_default_context()
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE
        return ctx

    async def login(self, username, password):
        if not self._protocol:
            if self._proxy_url:
                try:
                    import python_socks.async_.asyncio  # isort:skip
                    proxy = python_socks.async_.asyncio.Proxy.from_url(self._proxy_url.with_auth)
                except ValueError as e:
                    raise _errors.ValueError(e)

                try:
                    sock = await proxy.connect(
                        dest_host=self._host,
                        dest_port=self._port,
                        # Timeouts are handled with async_timeout in RPCBase
                        timeout=float('inf'),
                    )
                except python_socks.ProxyError as e:
                    raise _errors.ConnectionError(e)

                create_connection_kwargs = {
                    'sock': sock,
                    'server_hostname': self._host,
                    'protocol_factory': self._protocol_factory,
                    'ssl': self._create_ssl_context(),
                    # Timeouts are handled with async_timeout in RPCBase
                    'ssl_handshake_timeout': float('inf'),
                }
            else:
                create_connection_kwargs = {
                    'host': self._host,
                    'port': self._port,
                    'protocol_factory': self._protocol_factory,
                    'ssl': self._create_ssl_context(),
                    # Timeouts are handled with async_timeout in RPCBase
                    'ssl_handshake_timeout': float('inf'),
                }

            transport_, self._protocol = await _utils.catch_connection_exceptions(
                self._loop.create_connection(**create_connection_kwargs)
            )

            # The server expects a client_version kwarg when calling
            # daemon.login. We pretend to be the official client in case the
            # server attempts to derive capabilities from the client_version.
            daemon_version = await self.call('daemon.info')
            _log.debug('Logging in as %r', username)
            auth_level_ = await self.call(
                'daemon.login',
                username,
                password,
                client_version=daemon_version,
            )
            _log.debug('auth level: %r', auth_level_)

    async def logout(self):
        if self._protocol:
            _log.debug('Logging out')
            self._protocol.close()

    async def call(self, method, *args, **kwargs):
        request = _DelugeRPCRequest(
            method=method,
            args=args,
            kwargs=kwargs,
        )
        return await self._protocol.send_request(request)


class _DelugeRPCProtocol(asyncio.Protocol):
    PROTOCOL_VERSION = 1
    # See https://docs.python.org/3/library/struct.html#format-strings
    MESSAGE_HEADER_FORMAT = '!BI'
    MESSAGE_HEADER_SIZE = struct.calcsize(MESSAGE_HEADER_FORMAT)

    RPC_RESPONSE = 1
    RPC_ERROR = 2
    RPC_EVENT = 3

    def __init__(
            self,
            on_connection_made=None,
            on_connection_lost=None,
            on_event_received=None,
    ):
        self._on_connection_made = on_connection_made
        self._on_connection_lost = on_connection_lost
        self._on_event_received = on_event_received
        self._reset_internal_state()

    def _reset_internal_state(self):
        self._transport = None
        self._buffer = b''
        self._message_length = 0
        self._requests = {}
        _DelugeRPCRequest.reset_id()

    def connection_made(self, transport):
        _log.debug('Made connection: %r', transport)
        self._transport = transport
        if self._on_connection_made:
            self._on_connection_made()

    def connection_lost(self, exception):
        _log.debug('Lost connection: %r', exception)

        # Don't leave any ongoing requests hanging
        if self._requests:
            request_exception = exception or _errors.ConnectionError('Connection lost')
            _log.debug('Reporting exception to %d ongoing requests: %r',
                       len(self._requests), request_exception)
            for request in self._requests.values():
                if not request.future.done():
                    request.future.set_exception(request_exception)
            self._requests.clear()

        self.close()
        self._reset_internal_state()

        if self._on_connection_lost:
            self._on_connection_lost()

    def close(self):
        if self._transport:
            self._transport.close()

    def data_received(self, data):
        self._buffer += data

        while len(self._buffer) >= self.MESSAGE_HEADER_SIZE:
            if self._message_length == 0:
                self._handle_new_message()

            if len(self._buffer) >= self._message_length:
                self._handle_complete_message()
            else:
                break

    def _handle_new_message(self):
        # Read header bytes
        header = self._buffer[:self.MESSAGE_HEADER_SIZE]
        # Remove the header from the buffer
        self._buffer = self._buffer[self.MESSAGE_HEADER_SIZE:]
        # Unpack header bytes into usable objects
        protocol_version, self._message_length = struct.unpack(self.MESSAGE_HEADER_FORMAT, header)
        if protocol_version != self.PROTOCOL_VERSION:
            raise RuntimeError(f'Unsupported protocol version: {protocol_version}')

    def _handle_complete_message(self):
        # Consume message from buffer
        data = self._buffer[:self._message_length]
        self._buffer = self._buffer[self._message_length:]
        self._message_length = 0

        # Decode message
        import rencode, zlib  # isort:skip
        msg = rencode.loads(zlib.decompress(data), decode_utf8=True)
        msg_type = msg[0]

        # Handle message
        if msg_type == self.RPC_EVENT:
            event = msg[1]
            args = msg[2]
            _log.debug('RPC_EVENT: %s(%s)', event, ', '.join(repr(arg) for arg in args))
            if self._on_event_received:
                self._on_event_received(event, args)

        elif msg_type == self.RPC_ERROR:
            _log.debug('RPC_ERROR: %r', msg)
            request_id, exc_clsname, exc_posargs, exc_kwargs, traceback = msg[1:6]
            exc = self._create_rpc_error(exc_clsname, exc_posargs, exc_kwargs, traceback)
            self._set_response(request_id, exc)

        elif msg_type == self.RPC_RESPONSE:
            _log.debug('RPC_RESPONSE: %r', msg)
            request_id, response = msg[1:]
            self._set_response(request_id, response)

        else:
            raise RuntimeError(f'Unknown RPC message type: {msg_type!r}')

    def _create_rpc_error(self, clsname, posargs, kwargs, traceback):
        if clsname == 'BadLoginError':
            exc = _errors.AuthenticationError('Authentication failed')
        elif clsname == 'NotAuthorizedError':
            msg = 'Not authorized'
            if len(posargs) >= 2:
                auth_level, auth_level_required = posargs[:2]
                msg += f': Your authorization level is {auth_level}, but you need {auth_level_required}'
            exc = _errors.AuthenticationError(msg)
        elif posargs and posargs[0]:
            exc = _errors.RPCError(posargs[0])
        elif clsname:
            exc = _errors.RPCError(clsname)
        else:
            exc = _errors.RPCError('Unknown error')
        _log.debug('Server traceback for %r:\n%s', exc, traceback)
        return exc

    def _set_response(self, request_id, response):
        _log.debug('Setting response for request #%s: %r', request_id, response)
        try:
            request = self._requests[request_id]
        except KeyError:
            raise RuntimeError(f'Got response to unknown request #{request_id}: {response!r}')

        try:
            if request.future.done():
                raise RuntimeError(f'Request #{request_id} already has a response')
            else:
                if isinstance(response, BaseException):
                    request.future.set_exception(response)
                else:
                    request.future.set_result(response)
        finally:
            del self._requests[request_id]

    def _transfer_messages(self, *data):
        import rencode, zlib  # isort:skip
        body = zlib.compress(rencode.dumps(_utils.convert_to_basic_type(data)))
        body_len = len(body)
        packed = struct.pack(
            f'{self.MESSAGE_HEADER_FORMAT}{body_len}s',
            self.PROTOCOL_VERSION,
            body_len,
            body,
        )
        self._transport.write(packed)

    _valid_types = {
        list,
        tuple,
        dict,
        str,
        bytes,
        int,
        float,
        bool,
        type(None),
    }

    def _convert_to_valid_type(self, data):
        # rencode only allows specific types. Here we convert subclasses of
        # these types back to allowed types.

        data_type = type(data)
        if data_type == list:
            return [
                self._convert_to_valid_type(item)
                for item in data
            ]
        elif data_type == tuple:
            return tuple(
                self._convert_to_valid_type(item)
                for item in data
            )
        elif data_type == dict:
            return {
                self._convert_to_valid_type(k) : self._convert_to_valid_type(v)
                for k, v in data.items()
            }
        elif data_type in self._valid_types:
            return data

        for t in self._valid_types:
            if isinstance(data, t):
                return self._convert_to_valid_type(t(data))

        raise TypeError(f'Unsupported type: {type(data).__name__}: {data!r}')

    def send_request(self, request):
        _log.debug('Sending request #%d: %r', request.id, request)
        self._requests[request.id] = request
        self._transfer_messages(request.format_message())
        return request.future


class _DelugeRPCRequest:
    _next_id = 0

    def __init__(self, *, method, args, kwargs):
        self.method = str(method)
        self.args = tuple(args)
        self.kwargs = dict(kwargs)
        self.future = asyncio.Future()
        self.id = type(self)._next_id
        type(self)._next_id += 1

    @classmethod
    def reset_id(cls):
        """Reset automatically incremented :attr:`id` to ``0``"""
        cls._next_id = 0

    def format_message(self):
        """Return properly formatted request"""
        return (self.id, self.method, self.args, self.kwargs)

    def __repr__(self):
        text = f'{self.method}('
        posargs = ', '.join((repr(arg) for arg in self.args))
        kwargs = ', '.join((f'{k}={v!r}' for k, v in self.kwargs.items()))
        args = ', '.join((x for x in (posargs, kwargs) if x))
        if args:
            text += args
        text += ')'
        return text

    def __eq__(self, other):
        return (
            type(other) is type(self)
            and self.method == other.method
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    def __ne__(self, other):
        return not self.__eq__(other)



import asyncio
import inspect
import types
from unittest.mock import Mock, NonCallableMock


# AsyncMock was added in Python 3.8
class AsyncMock(Mock):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Make asyncio.iscoroutine[function]() work
        code_mock = NonCallableMock(spec_set=types.CodeType)
        code_mock.co_flags = inspect.CO_COROUTINE
        self.__dict__['__code__'] = code_mock
        self.__dict__['_is_coroutine'] = asyncio.coroutines._is_coroutine

    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)


def make_url_parts(url):
    return {
        name: getattr(url, name)
        for name in ('scheme', 'host', 'port', 'path', 'username', 'password')
    }



import abc
import asyncio
import collections
import weakref

import async_timeout

from . import _errors, _utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class RPCBase(abc.ABC):
    """Base class for BitTorrent client RPC interfaces"""

    # Abstract methods

    @abc.abstractmethod
    async def _connect(self):
        """Connect to RPC interface"""

    @abc.abstractmethod
    async def _disconnect(self):
        """
        Disconnect from RPC interface

        This method should not raise an exception if it fails. For example, if
        the service at :attr:`url` is down, we can't logout, which shouldn't be
        an error.
        """

    @abc.abstractmethod
    async def _call(self):
        """Call RPC method"""

    # Abstract properties

    @property
    @abc.abstractmethod
    def name(self):
        """Lowercase name of the BitTorrent client"""

    @property
    @abc.abstractmethod
    def label(self):
        """Properly capitalized :attr:`name` of the BitTorrent client"""

    # Properties

    default_timeout = 60.0
    """Default :attr:`timeout`"""

    @property
    def timeout(self):
        """
        Timeout in seconds for RPC calls

        If this property is set to a falsy value, :attr:`default_timeout` is
        used.

        :raise ValueError: if set to something that can't be coerced into a
            :class:`float`
        """
        return getattr(self, '_timeout', self.default_timeout)

    @timeout.setter
    def timeout(self, timeout):
        try:
            self._timeout = float(timeout) if timeout else self.default_timeout
        except (TypeError, ValueError):
            raise _errors.ValueError('Invalid timeout')
        else:
            self._invalidate_http_client()

    @property
    @abc.abstractmethod
    def URL(self):
        """
        :class:`~.URL` subclass

        Setting properties on an instance of this subclass to invalid values
        should raise :class:`~.ValueError`. For example, :class:`~.RtorrentURL`
        only accepts the schemes ``http``, ``scgi`` or ``file``.
        """

    @property
    def url(self):
        """
        :class:`~.URL` of the RPC interface

        Changing any of the properties will re-connect to the new URL on the
        next :meth:`~.call`.

        If this property is set to a falsy value, :attr:`~.URL.default` is
        used.

        :raise ValueError: if set to an invalid URL
        """
        url = getattr(self, '_url', None)
        if not url:
            url = self._url = type(self).URL(
                url=type(self).URL.default,
                on_change=self._invalidate_http_client,
            )
        return url

    @url.setter
    def url(self, url):
        self._url = type(self).URL(
            url=url if url else type(self).URL.default,
            on_change=self._invalidate_http_client,
        )
        self._invalidate_http_client()

    @property
    def proxy_url(self):
        """
        SOCKS5, SOCKS4 or HTTP proxy :class:`~.URL` for tunneling the RPC connection

        If this property is set to a falsy value, no proxy is used.

        :raise ValueError: if set to an invalid URL
        """
        return getattr(self, '_proxy_url', None)

    @proxy_url.setter
    def proxy_url(self, proxy_url):
        if proxy_url:
            self._proxy_url = _utils.URL(
                url=proxy_url,
                on_change=self._invalidate_http_client,
            )
        else:
            self._proxy_url = None
        self._invalidate_http_client()

    @property
    def status(self):
        """:class:`~.ConnectionStatus` enum"""
        return getattr(self, '_status', _utils.ConnectionStatus.disconnected)

    @property
    def is_connected(self):
        """
        `True` if :attr:`status: is :attr:`~.ConnectionStatus.connected`, `False`
        otherwise
        """
        return self.status is _utils.ConnectionStatus.connected

    # Callbacks

    def set_connecting_callback(self, callback, *args, **kwargs):
        """
        Schedule callback to be called when a connection attempt is made

        :param callback: Callable to call

        All remaining arguments are passed to `callback` when it is called.
        """
        assert callable(callback)
        self._connection_callbacks['connecting'].append((callback, args, kwargs))

    def unset_connecting_callback(self, callback):
        """
        Unschedule callback to be called when a connection attempt is made

        :param callback: Callable to call
        """
        assert callable(callback)
        for cb in tuple(self._connection_callbacks['connecting']):
            if cb[0] == callback:
                self._connection_callbacks['connecting'].remove(cb)

    def set_connected_callback(self, callback, *args, **kwargs):
        """
        Schedule callback to be called when a connection attempt was successful

        :param callback: Callable to call

        All remaining arguments are passed to `callback` when it is called.
        """
        assert callable(callback)
        self._connection_callbacks['connected'].append((callback, args, kwargs))

    def unset_connected_callback(self, callback):
        """
        Unschedule callback to be called when a connection attempt was successful

        :param callback: Callable to call
        """
        assert callable(callback)
        for cb in tuple(self._connection_callbacks['connected']):
            if cb[0] == callback:
                self._connection_callbacks['connected'].remove(cb)

    def set_disconnected_callback(self, callback, *args, **kwargs):
        """
        Schedule callback to be called when the connection was lost or a connection
        attempt has failed

        :param callback: Callable to call

        All remaining arguments are passed to `callback` when it is called.
        """
        assert callable(callback)
        self._connection_callbacks['disconnected'].append((callback, args, kwargs))

    def unset_disconnected_callback(self, callback):
        """
        Unschedule callback to be called when the connection was lost or a
        connection attempt has failed

        :param callback: Callable to call
        """
        assert callable(callback)
        for cb in tuple(self._connection_callbacks['disconnected']):
            if cb[0] == callback:
                self._connection_callbacks['disconnected'].remove(cb)

    @_utils.cached_property
    def _connection_callbacks(self):
        return {
            'connecting': [],
            'connected': [],
            'disconnected': [],
        }

    def _call_connection_callbacks(self, name):
        for callback, args, kwargs in self._connection_callbacks[name]:
            callback(*args, **kwargs)

    # RPC methods

    @_utils.cached_property
    def _connection_lock(self):
        return asyncio.Lock()

    async def connect(self):
        """
        Connect to RPC interface

        Do nothing if :attr:`status` indicates we are already connected.

        It is safe to call this method multiple times concurrently. The first
        call will actually connect while the remaining calls wait for the first
        call to finish. If the first call fails, each of the remaining calls
        will become the first call, i.e. it will attempt to connect while the
        others wait for it.

        :raise ConnectionError: if the request failed
        :raise AuthenticationError: if authentication failed
        :raise TimeoutError: if there is no response after :attr:`timeout` seconds
        :raise RPCError: if there is any miscommunication between us and the RPC
            interface
        """
        _log.debug('%s: connect(): Waiting for connection lock (status=%s)', self.label, self.status)
        try:
            async with self._connection_lock:
                _log.debug('%s: connect(): Acquired connection lock (status=%s)', self.label, self.status)
                if self.status is not _utils.ConnectionStatus.connected:
                    self._status = _utils.ConnectionStatus.connecting
                    self._call_connection_callbacks('connecting')
                    try:
                        async with async_timeout.timeout(self.timeout):
                            await self._connect()

                    except Exception as e:
                        _log.debug('%s: Failed to connect: %r', self.label, e)
                        await self._disconnect()
                        self._status = _utils.ConnectionStatus.disconnected
                        self._call_connection_callbacks('disconnected')
                        if isinstance(e, asyncio.TimeoutError):
                            raise _errors.TimeoutError(f'Timeout after {self.timeout} seconds')
                        else:
                            raise

                    else:
                        _log.debug('%s: Connected', self.label)
                        self._status = _utils.ConnectionStatus.connected
                        self._call_connection_callbacks('connected')

        finally:
            _log.debug('%s: connect(): Freed connection lock (status=%s)', self.label, self.status)

    async def disconnect(self):
        """
        Disconnect from RPC interface

        Do nothing if :attr:`status` indicates we are already disconnected.

        It is safe to call this method multiple times concurrently. The first
        call will actually disconnect while the remaining calls wait for the
        first call to finish. If the first call fails, each of the remaining
        call will become the first call, i.e. it will attempt to disconnect
        while the others wait for it.

        :attr:`status` is always :attr:`.ConnectionStatus.disconnected` when
        this method returns, regardless of any raised exceptions.

        :raise AuthenticationError: if authentication failed
        :raise TimeoutError: if there is no response after :attr:`timeout` seconds
        :raise RPCError: if there is any miscommunication between us and the RPC
            interface
        """
        _log.debug('%s: disconnect(): Waiting for connection lock (status=%s)', self.label, self.status)
        try:
            async with self._connection_lock:
                _log.debug('%s: disconnect(): Acquired connection lock (status=%s)', self.label, self.status)
                if self.status is not _utils.ConnectionStatus.disconnected:
                    try:
                        async with async_timeout.timeout(self.timeout):
                            await self._disconnect()

                    except asyncio.TimeoutError:
                        _log.debug('%s: disconnect(): Timeout', self.label)
                        raise _errors.TimeoutError(f'Timeout after {self.timeout} seconds')

                    finally:
                        self._status = _utils.ConnectionStatus.disconnected
                        self._call_connection_callbacks('disconnected')

        finally:
            await self._close_http_client()
            _log.debug('%s: disconnect(): Freed connection lock (status=%s)', self.label, self.status)

    async def call(self, *args, **kwargs):
        """
        Call RPC method and return the result

        If :attr:`status` is not :attr:`.ConnectionStatus.connected`, call
        :meth:`connect` first.

        :raise AuthenticationError: if authentication failed
        :raise ConnectionError: if the request failed
        :raise TimeoutError: if there is no response after :attr:`timeout` seconds
        :raise RPCError: if there is any miscommunication between us and the RPC
            interface

        :return: the return value of the RPC method

            This should be decoded bytes, deserialized JSON, etc. Exceptions
            from decoding and deserializing should be raised as
            :class:`~.RPCError`.
        """
        # _log.debug('%s: [%s] Calling: %s, %s', self.label, self.status, args, kwargs)
        if self.status is not _utils.ConnectionStatus.connected:
            _log.debug('%s: Auto-connecting', self.label)
            await self.connect()

        try:
            async with async_timeout.timeout(self.timeout):
                return await self._call(*args, **kwargs)

        except asyncio.TimeoutError:
            _log.debug('%s: call(): Timeout', self.label)
            raise _errors.TimeoutError(f'Timeout after {self.timeout} seconds')

    # Events

    @_utils.cached_property
    def _event_handlers(self):
        return collections.defaultdict(lambda: [])

    async def add_event_handler(self, event, handler, autoremove=False):
        """
        Call callable when event happens

        :param event: Name or other identifier of the event (refer to the client
            documentation for valid values)
        :param handler: Callable to be called when `event` happens (refer to the
            client documentation for the call signature)
        :param autoremove: Whether `handler` should be unregistered
            automatically after it is garbage collected (see :mod:`weakref`)

        If `handler` is already registered for `event`, do nothing.

        Multiple handlers can be registered for the same event. They will be
        called in the order they are added.

        :raise NotImplementedError: if the client doesn't support events
        """
        assert callable(handler), handler

        if event not in self._event_handlers:
            await self._subscribe(event)

        if autoremove:
            handler_ref = weakref.ref(handler, self._event_handlers[event].remove)
            if handler_ref not in self._event_handlers[event]:
                self._event_handlers[event].append(handler_ref)
                _log.debug('Added handler reference for event %r: %r', event, handler_ref)

        else:
            if handler not in self._event_handlers[event]:
                self._event_handlers[event].append(handler)
                _log.debug('Added handler for event %r: %r', event, handler)

    async def remove_event_handler(self, event, handler):
        """
        Stop calling callable when event happens

        See :meth:`add_event_handler`.

        If `handler` is not registered for `event`, do nothing.

        :raise NotImplementedError: if the client doesn't support events
        """
        event_handlers = self._event_handlers[event]

        # Disconnect `handler` from `event`
        if handler in event_handlers:
            event_handlers.remove(handler)

        # If there aren't any other handlers for `event`, unsubscribe
        if not event_handlers:
            del self._event_handlers[event]
            await self._unsubscribe(event)
            _log.debug('Removed handler for event %r: %r', event, handler)

    async def wait_for_event(self, event):
        """
        Block until event happens

        :param event: Name or other identifier of the event (refer to the client
            documentation for valid values)

        :raise NotImplementedError: if the client doesn't support events
        """
        blocker = asyncio.Event()

        def unblock_blocker(*_, **__):
            blocker.set()

        await self.add_event_handler(event, unblock_blocker)

        try:
            _log.debug('Waiting for event: %r', event)
            await blocker.wait()
            _log.debug('Done waiting for event: %r', event)
        finally:
            await self.remove_event_handler(event, unblock_blocker)

    async def _emit_event(self, event, args=None, kwargs=None):
        # This function is called by subclasses when they receive an event
        args = args or ()
        kwargs = kwargs or {}
        for handler in self._event_handlers[event]:
            # `handler` is a weakref.ref instance if it was registered with
            # `autoremove=True`.
            #
            # NOTE: If we pass `handler` to _log.debug() here, it will keep
            #       another strong reference to `handler` and the test that
            #       covers dead weakrefs will fail.
            if isinstance(handler, weakref.ref):
                handler = handler()
                if handler is None:
                    continue

            # Because this method is likely being called by some background task
            # that is waiting for events, any exception raised by `handler` will
            # go unnoticed, so we stop the loop if that happens.
            # There should be a better way to handle this, but I can't find it.
            try:
                if asyncio.iscoroutinefunction(handler):
                    await handler(*args, **kwargs)
                else:
                    handler(*args, **kwargs)

            except BaseException as e:
                try:
                    loop = asyncio.get_running_loop()
                except RuntimeError:
                    # Loop is already stopped
                    pass
                else:
                    _log.debug('Stopping loop because event handler for %r '
                               'threw an exception: %r', event, e)
                    loop.stop()

                # We probably raise this into the void, but not raising
                # BaseException is really bad.
                raise

    async def _subscribe(self, event):
        # Tell the client to send us a certain type of event
        raise NotImplementedError(f'Events are not supported for {self.label}')

    async def _unsubscribe(self, event):
        # Tell the client to stop sending us a certain type of event
        raise NotImplementedError(f'Events are not supported for {self.label}')

    # Context manager

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        _log.debug('%s: Disconnecting at end of context manager', self.label)
        await self.disconnect()
        # Make extra sure the HTTP client is closed
        await self._close_http_client()

    # HTTP client

    @property
    def _http_headers(self):
        """
        Headers that are sent on HTTP requests

        Subclass instances should modify these headers as they please. They are
        automatically combined with any obligatory headers.
        """
        headers = getattr(self, '__http_headers', None)
        if headers is None:
            headers = {}
            setattr(self, '__http_headers', headers)
        return headers

    async def _get_http_client(self):
        if getattr(self, '_http_client_is_invalidated', False):
            delattr(self, '_http_client_is_invalidated')
            _log.debug('%s: HTTP client was invalidated', self.label)
            await self._close_http_client()

        if not hasattr(self, '_http_client'):
            proxy_url = self.proxy_url.with_auth if self.proxy_url else None
            self._http_client = _utils.create_http_client(
                auth=(self.url.username, self.url.password),
                proxy_url=proxy_url,
            )
            _log.debug('%s: Created new HTTP client: %r', self.label, self._http_client)
        return self._http_client

    async def _close_http_client(self):
        if hasattr(self, '_http_client'):
            _log.debug('%s: Closing HTTP client: %r', self.label, self._http_client)
            await self._http_client.aclose()
            delattr(self, '_http_client')

    def _invalidate_http_client(self):
        if hasattr(self, '_http_client'):
            _log.debug('%s: HTTP client is now invalidated: %r', self.label, self._http_client)
            self._http_client_is_invalidated = True
        self._status = _utils.ConnectionStatus.disconnected

    async def _send_post_request(self, url, data=None, files=None):
        # httpx wants dictionaries as `data` and everything else as `content`
        if data is not None and not isinstance(data, dict):
            content = data
            data = None
        else:
            content = None

        client = await self._get_http_client()
        return await _utils.catch_connection_exceptions(
            client.post(
                timeout=float('inf'),
                url=url,
                headers=self._http_headers,
                data=data,
                content=content,
                files=files,
            ),
        )



"""
Asynchronous low-level communication with BitTorrent clients
"""

__project_name__ = 'aiobtclientrpc'
__description__ = 'Asynchronous low-level communication with BitTorrent clients'
__homepage__ = 'https://codeberg.org/plotski/aiobtclientrpc'
__version__ = '4.0.0'
__author__ = 'plotski'
__author_email__ = 'plotski@example.org'

# isort:skip_file

from ._base import RPCBase
from ._errors import *

from ._deluge import DelugeRPC, DelugeURL
from ._qbittorrent import QbittorrentRPC, QbittorrentURL
from ._rtorrent import RtorrentRPC, RtorrentURL
from ._transmission import TransmissionRPC, TransmissionURL

from ._utils import URL, ConnectionStatus, client, clients



import asyncio
import re
import sys
import xmlrpc
from unittest.mock import Mock, call

import pytest

from aiobtclientrpc import RPCBase, _errors, _rtorrent, _utils

from .common import AsyncMock, make_url_parts


@pytest.mark.parametrize(
    argnames='url, exp',
    argvalues=(
        # Default URL
        ('',
         {'scheme': 'scgi', 'host': '127.0.0.1', 'port': '5000', 'path': None, 'username': None, 'password': None}),

        # RPC socket path
        ('/abs/path/to/rpc.socket',
         {'scheme': 'file', 'host': None, 'port': None, 'path': '/abs/path/to/rpc.socket', 'username': None, 'password': None}),
        ('file:///path/to/rpc.socket',
         {'scheme': 'file', 'host': None, 'port': None, 'path': '/path/to/rpc.socket', 'username': None, 'password': None}),
        ('file://foo:bar@path/to/rpc.socket',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'foo:bar@path/to/rpc.socket', 'username': None, 'password': None}),
        ('file://path:123/to/rpc.socket',
         {'scheme': 'file', 'host': None, 'port': None, 'path': 'path:123/to/rpc.socket', 'username': None, 'password': None}),

        # SCGI
        ('myhost',
         {'scheme': 'scgi', 'host': 'myhost', 'port': '5000', 'path': None, 'username': None, 'password': None}),
        ('myhost:123',
         {'scheme': 'scgi', 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('scgi://myhost:123',
         {'scheme': 'scgi', 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),

        # HTTP(s)
        ('http://myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': None, 'username': None, 'password': None}),
        ('http://foo:bar@myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('https://foo:bar@myhost:123/RPC123',
         {'scheme': 'https', 'host': 'myhost', 'port': '123', 'path': '/RPC123', 'username': 'foo', 'password': 'bar'}),
        ('https://foo:bar@myhost',
         {'scheme': 'https', 'host': 'myhost', 'port': '5000', 'path': None, 'username': 'foo', 'password': 'bar'}),
        ('https://myhost/RPC/17',
         {'scheme': 'https', 'host': 'myhost', 'port': '5000', 'path': '/RPC/17', 'username': None, 'password': None}),

        # Invalid scheme
        ('arf://myhost',
         _errors.ValueError('Scheme must be "file", "scgi", "http" or "https"')),

        # Invalid scgi URL
        ('scgi://foo:@myhost',
         _errors.ValueError("scgi URLs don't have a username")),
        ('scgi://:foo@myhost',
         _errors.ValueError("scgi URLs don't have a password")),
        ('scgi://myhost/this/is/a/path',
         _errors.ValueError("scgi URLs don't have a path")),
    ),
    ids=lambda v: str(v),
)
def test_RtorrentURL(url, exp):
    if isinstance(exp, Exception):
        with pytest.raises(type(exp), match=rf'^{re.escape(str(exp))}$'):
            _rtorrent.RtorrentURL(url)
    else:
        url = _rtorrent.RtorrentURL(url)
        assert make_url_parts(url) == exp


class AsyncIterator:
    def __init__(self, items):
        self._items = list(items)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self._items:
            return self._items.pop(0)
        else:
            raise StopAsyncIteration


@pytest.mark.parametrize('url', (None, 'http://a:b@foo:123'))
@pytest.mark.parametrize(
    argnames='kwargs',
    argvalues=(
        {},
        {'scheme': 'https'},
        {'host': 'asdf'},
        {'port': '123'},
        {'scheme': 'http', 'username': 'this', 'password': 'that'},
        {'timeout': 123},
        {'proxy_url': 'http://hey:ho@bar:456'},
    ),
    ids=lambda v: str(v),
)
def test_RtorrentRPC_instantiation(kwargs, url):
    if url:
        kwargs['url'] = url
    rpc = _rtorrent.RtorrentRPC(**kwargs)

    default_url = _utils.URL(_rtorrent.RtorrentURL.default)
    exp_url = _utils.URL(_rtorrent.RtorrentURL.default)
    if url:
        custom_url = _utils.URL(url)
    for name in ('scheme', 'host', 'port', 'path', 'username', 'password'):
        if name in kwargs:
            exp_value = kwargs[name]
        elif url:
            exp_value = getattr(custom_url, name)
        else:
            exp_value = getattr(default_url, name)
        actual_value = getattr(rpc.url, name)
        assert actual_value == exp_value
        setattr(exp_url, name, exp_value)

    assert rpc.url == exp_url
    assert rpc.timeout == kwargs.get('timeout', RPCBase.default_timeout)
    if 'proxy_url' in kwargs:
        assert rpc.proxy_url == _utils.URL(kwargs['proxy_url'])
    else:
        assert rpc.proxy_url is None

@pytest.mark.parametrize(
    argnames='kwargs, exp_error',
    argvalues=(
        ({'url': 'bar:baz'}, 'Invalid port'),
        ({'port': (1, 2, 3)}, 'Invalid port'),
        ({'scheme': 'file', 'host': 'localhost'}, "file URLs don't have a host"),
        ({'timeout': 'never'}, 'Invalid timeout'),
        ({'proxy_url': 'foo://bar:baz'}, 'Invalid port'),
    ),
    ids=lambda v: str(v),
)
def test_RtorrentRPC_instantiation_with_invalid_argument(kwargs, exp_error):
    with pytest.raises(_errors.ValueError, match=rf'^{re.escape(exp_error)}$'):
        _rtorrent.RtorrentRPC(**kwargs)


@pytest.mark.parametrize('proxy_url', (None, 'http://a:b@proxy.local'))
@pytest.mark.asyncio
async def test_RtorrentRPC_connect(proxy_url, mocker):
    rpc = _rtorrent.RtorrentRPC(proxy_url=proxy_url)
    mocker.patch.object(rpc, '_disconnect', AsyncMock())
    AsyncServerProxy_mock = mocker.patch('aiobtclientrpc._rtorrent._AsyncServerProxy')
    mocker.patch.object(rpc, '_call', AsyncMock())

    await rpc._connect()

    assert rpc._disconnect.call_args_list == [call()]

    if proxy_url:
        assert AsyncServerProxy_mock.call_args_list == [call(
            url=rpc.url,
            proxy_url=rpc.proxy_url,
        )]
    else:
        assert AsyncServerProxy_mock.call_args_list == [call(
            url=rpc.url,
            proxy_url=None,
        )]

    assert rpc._call.call_args_list == [call('system.pid')]


@pytest.mark.parametrize('xmlrpc', (None, AsyncMock()))
@pytest.mark.asyncio
async def test_RtorrentRPC_disconnect(xmlrpc):
    rpc = _rtorrent.RtorrentRPC()
    if xmlrpc is not None:
        rpc._xmlrpc = xmlrpc

    await rpc._disconnect()

    if xmlrpc:
        assert xmlrpc.close.call_args_list == [call()]

    assert not hasattr(rpc, '_xmlrpc')


@pytest.mark.parametrize(
    argnames='raised_exception, exp_exception',
    argvalues=(
        (
            xmlrpc.client.ProtocolError('http://url', 401, 'Something went wrong', {'head': 'ers'}),
            _errors.AuthenticationError('Authentication failed'),
        ),
        (
            xmlrpc.client.ProtocolError('http://url', 123, 'Something went wrong', {'head': 'ers'}),
            _errors.RPCError('Something went wrong'),
        ),
        (
            xmlrpc.client.Fault(123, 'Something else went wrong'),
            _errors.RPCError('Something else went wrong'),
        ),
        (
            None,
            None,
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_RtorrentRPC_call(raised_exception, exp_exception, mocker):
    rpc = _rtorrent.RtorrentRPC()
    rpc._xmlrpc = Mock(call=Mock())
    catch_connection_exceptions_mock = mocker.patch('aiobtclientrpc._utils.catch_connection_exceptions', AsyncMock())
    if raised_exception:
        catch_connection_exceptions_mock.side_effect = raised_exception

    method = 'some_method'
    args = ('foo', 'bar', 'baz')

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._call(method, *args)
    else:
        return_value = await rpc._call(method, *args)
        assert return_value is catch_connection_exceptions_mock.return_value

    assert catch_connection_exceptions_mock.call_args_list == [call(rpc._xmlrpc.call.return_value)]
    assert rpc._xmlrpc.call.call_args_list == [call(method, *args)]


@pytest.mark.parametrize(
    argnames='calls, responses, kwargs, exp_result, exp_system_multicall_called',
    argvalues=(
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456'), ('foo', 'f00')],
            [['first response'], {'faultString': 'Bad!'}, [789], ['00']],
            {'raise_errors': False},
            ['first response', _errors.RPCError('Bad!'), 789, '00'],
            True,
        ),
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456'), ('foo', 'f00')],
            [['first response'], [['b', 'a', 'r']], {'faultString': 'Bad!'}, ['00']],
            {'raise_errors': True},
            _errors.RPCError('Bad!'),
            True,
        ),
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456'), ('foo', 'f00')],
            [['first response'], {'faultString': 'Bad!'}, [789], ['00']],
            {'raise_errors': False, 'as_dict': True},
            RuntimeError("Multiple foo calls: foo('a', 'b'), foo('f00')"),
            False,
        ),
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456')],
            [['first response'], {'faultString': 'Bad!'}, [789]],
            {'raise_errors': False, 'as_dict': True},
            {'foo': 'first response', 'bar': _errors.RPCError('Bad!'), 'baz': 789},
            True,
        ),
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456')],
            [['first response'], {'faultString': 'Bad!'}, [789]],
            {'raise_errors': True, 'as_dict': True},
            _errors.RPCError('Bad!'),
            True,
        ),
        (
            [('foo', 'a', 'b'), ['bar'], ('baz', 123, '456')],
            [['first response'], 'unexpected return value', [789]],
            {},
            RuntimeError("Unexpected response: 'unexpected return value'"),
            True,
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_RtorrentRPC_multicall(calls, responses, kwargs, exp_result, exp_system_multicall_called, mocker):
    rpc = _rtorrent.RtorrentRPC()
    mocker.patch.object(rpc, 'call', AsyncMock(return_value=responses))

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await rpc.multicall(*calls, **kwargs)
    else:
        return_value = await rpc.multicall(*calls, **kwargs)
        assert return_value == exp_result

    if exp_system_multicall_called:
        assert rpc.call.call_args_list == [call('system.multicall', [
            {'methodName': method, 'params': params}
            for method, *params in calls
        ])]
    else:
        assert rpc.call.call_args_list == []


@pytest.mark.parametrize(
    argnames='supported_methods, candidates, exp_method, exp_exception',
    argvalues=(
        (
            ('foo', 'bar', 'baz'),
            ('bar', 'baz'),
            'bar',
            None,
        ),
        (
            ('foo', 'bar', 'baz'),
            ('asdf', 'f00'),
            None,
            _errors.ValueError("Unsupported method(s): 'asdf', 'f00'"),
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_RtorrentRPC_get_supported_method(supported_methods, candidates, exp_method, exp_exception, mocker):
    rpc = _rtorrent.RtorrentRPC()
    mocker.patch.object(rpc, 'call', AsyncMock(return_value=supported_methods))

    if not type(rpc)._supported_methods:
        exp_calls = [call('system.listMethods')]
    else:
        exp_calls = []

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc.get_supported_method(*candidates)
    else:
        # Multiple get_supported_method() calls, but only one "system.list_method" request
        for _ in range(3):
            return_value = await rpc.get_supported_method(*candidates)
            assert return_value == exp_method

    assert rpc.call.call_args_list == exp_calls


@pytest.mark.parametrize('proxy_url', (None, _utils.URL('http://proxy.local')))
@pytest.mark.parametrize(
    argnames='url, exp_transport, exp_exception',
    argvalues=(
        (_utils.URL('http://foo.local:123'), '_HttpTransport', None),
        (_utils.URL('https://foo.local:123'), '_HttpTransport', None),
        (_utils.URL('scgi://foo.local:123'), '_ScgiHostTransport', None),
        (_utils.URL('file://path/to/socket'), '_ScgiSocketTransport', None),
        (_utils.URL('foo://bar.baz'), None, _errors.ValueError('Unsupported protocol: foo://bar.baz')),
    ),
    ids=lambda v: str(v),
)
def test_AsyncServerProxy(url, exp_transport, exp_exception, proxy_url, mocker):
    transport_mocks = {
        '_HttpTransport': mocker.patch('aiobtclientrpc._rtorrent._HttpTransport'),
        '_ScgiHostTransport': mocker.patch('aiobtclientrpc._rtorrent._ScgiHostTransport'),
        '_ScgiSocketTransport': mocker.patch('aiobtclientrpc._rtorrent._ScgiSocketTransport'),
    }

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            _rtorrent._AsyncServerProxy(url, proxy_url)
    else:
        if str(url).startswith('file://') and proxy_url:
            with pytest.raises(_errors.ValueError, match=rf'^You cannot use a proxy to connect to {url}$'):
                _rtorrent._AsyncServerProxy(url, proxy_url)
        else:
            proxy = _rtorrent._AsyncServerProxy(url, proxy_url)
            assert proxy._transport is transport_mocks[exp_transport].return_value

            if str(url).startswith('file://'):
                assert transport_mocks[exp_transport].call_args_list == [
                    call(url=url),
                ]
            elif proxy_url:
                assert transport_mocks[exp_transport].call_args_list == [
                    call(url=url, proxy_url=proxy_url),
                ]
            else:
                assert transport_mocks[exp_transport].call_args_list == [
                    call(url=url, proxy_url=None),
                ]


@pytest.mark.asyncio
async def test_AsyncServerProxy_call(mocker):
    proxy = _rtorrent._AsyncServerProxy(_utils.URL('http://foo.baz'))

    dumps_mock = mocker.patch('xmlrpc.client.dumps')
    convert_to_basic_type_mock = mocker.patch('aiobtclientrpc._utils.convert_to_basic_type')
    mocker.patch.object(proxy._transport, 'request', Mock())
    mocker.patch.object(proxy, '_parse_response', AsyncMock())

    return_value = await proxy.call('foo', 'bar', 'baz')
    assert return_value is proxy._parse_response.return_value
    assert proxy._parse_response.call_args_list == [call(proxy._transport.request.return_value)]
    assert proxy._transport.request.call_args_list == [call(dumps_mock.return_value.encode.return_value)]
    assert dumps_mock.call_args_list == [call(
        convert_to_basic_type_mock.return_value,
        'foo',
        encoding='utf-8',
        allow_none=False,
    )]
    assert dumps_mock.return_value.encode.call_args_list == [call('utf-8', 'xmlcharrefreplace')]


@pytest.mark.parametrize(
    argnames='u_close_return_value, exp_return_value',
    argvalues=(
        (['single return value'], 'single return value'),
        (['multiple', 'return values'], ['multiple', 'return values']),
    ),
)
@pytest.mark.asyncio
async def test_AsyncServerProxy_parse_response(u_close_return_value, exp_return_value, mocker):
    proxy = _rtorrent._AsyncServerProxy(_utils.URL('http://foo.baz'))

    mocks = Mock()
    mocks.u.close.return_value = u_close_return_value
    mocker.patch('xmlrpc.client.getparser', return_value=(mocks.p, mocks.u))

    async def chunk_generator():
        for chunk in ('foo', 'bar', 'baz'):
            yield chunk

    chunks = chunk_generator()
    return_value = await proxy._parse_response(chunks)
    assert return_value == exp_return_value
    assert mocks.mock_calls == [
        call.p.feed('foo'),
        call.p.feed('bar'),
        call.p.feed('baz'),
        call.p.close(),
        call.u.close(),
    ]


@pytest.mark.asyncio
async def test_AsyncServerProxy_close(mocker):
    proxy = _rtorrent._AsyncServerProxy(_utils.URL('http://foo.baz'))
    mocker.patch.object(proxy._transport, 'close', AsyncMock())
    assert proxy._transport.close.call_args_list == []
    await proxy.close()
    assert proxy._transport.close.call_args_list == [call()]


@pytest.mark.parametrize(
    argnames='url, proxy_url, exp_url, exp_exception',
    argvalues=(
        (_utils.URL('file://path/to/proxy'), None, None, _errors.ValueError('Unsupported protocol: file')),
        (_utils.URL('http://a:b@foo.bar:123'), None, 'http://foo.bar:123/RPC2', None),
        (_utils.URL('http://a:b@foo.bar:123/custom/path'), None, 'http://foo.bar:123/custom/path', None),
        (_utils.URL('http://a:b@foo.bar/path'), _utils.URL('http://a:b@proxy'), 'http://foo.bar/path', None),
        (_utils.URL('https://foo.bar'), _utils.URL('https://proxy'), 'https://foo.bar/RPC2', None),
    ),
)
def test_HttpTransport(url, proxy_url, exp_url, exp_exception, mocker):
    create_http_client_mock = mocker.patch('aiobtclientrpc._utils.create_http_client', Mock())

    url = _utils.URL(url)
    if proxy_url:
        proxy_url = _utils.URL(proxy_url)

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            _rtorrent._HttpTransport(url, proxy_url=proxy_url)
    else:
        transport = _rtorrent._HttpTransport(url, proxy_url=proxy_url)
        assert transport._url == exp_url
        assert transport._http_client is create_http_client_mock.return_value
        assert isinstance(transport._request_lock, asyncio.Lock)

        if proxy_url:
            assert create_http_client_mock.call_args_list == [call(
                auth=(url.username, url.password),
                proxy_url=proxy_url.with_auth,
            )]
        else:
            assert create_http_client_mock.call_args_list == [call(
                auth=(url.username, url.password),
                proxy_url=None,
            )]


@pytest.mark.asyncio
async def test_HttpTransport_request_and_close(mocker):
    transport = _rtorrent._HttpTransport(_utils.URL('http://foo'))
    chunks = ('a', 'b', 'c')
    delays = (0.1, 0.2, 0.3)
    aiterator = AsyncIterator(chunks)
    mocks = Mock(
        aclose=AsyncMock(),
        _request=Mock(return_value=aiterator),
    )
    mocker.patch.object(transport._http_client, 'aclose', mocks.aclose)
    mocker.patch.object(transport, '_request', mocks._request)

    async def make_request(data):
        exp_chunks = list(chunks)
        my_delays = list(delays)
        async for x in transport.request(data):
            assert x == exp_chunks.pop(0)
            await asyncio.sleep(my_delays.pop(0))

    calls = [
        make_request('foo'),
        transport.close(),
        make_request('bar'),
        transport.close(),
        make_request('baz'),
        transport.close(),
    ]
    await asyncio.gather(*calls)

    # Even though each requested chunk is is received after an increasing delay,
    # and close() calls are not delayed at all, calls should be made in the same
    # order as they were dispatched.
    assert mocks.mock_calls == [
        call._request('foo'),
        call.aclose(),
        call._request('bar'),
        call.aclose(),
        call._request('baz'),
        call.aclose(),
    ]


@pytest.mark.parametrize(
    argnames='url, exp_url, status_code, reason_phrase, headers, raises',
    argvalues=(
        ('http://foo', 'http://foo/RPC2', 200, None, None, False),
        ('http://foo', 'http://foo/RPC2', 123, 'No such method', {'foo': 'bar'}, True),
    ),
)
@pytest.mark.asyncio
# I don't know how to mock async context managers in Python < 3.8.
@pytest.mark.skipif(sys.version_info < (3, 8), reason='requires Python 3.8 or higher')
async def test_HttpTransport_request(url, exp_url, status_code, reason_phrase, headers, raises, mocker):
    transport = _rtorrent._HttpTransport(_utils.URL(url))
    chunks = ('a', 'b', 'c')
    aiterator = AsyncIterator(chunks)
    response = Mock(
        aiter_bytes=Mock(return_value=aiterator),
        status_code=status_code,
        reason_phrase=reason_phrase,
        headers=headers,
    )

    # IMPORTANT: Tests pass without this, but any exception raised in
    #            HttpTransport._request() is ignored unless we explicitly raise
    #            it in __aexit__().
    async def __aexit__(self, exc_type=None, exc_value=None, traceback=None):
        if exc_value:
            raise exc_value

    mocker.patch.object(transport._http_client, 'stream', return_value=AsyncMock(
        __aenter__=AsyncMock(return_value=response),
        __aexit__=__aexit__,
    ))

    mock_data = 'mock data'
    exp_chunks = list(chunks)

    if raises:
        exp_exception = xmlrpc.client.ProtocolError(
            url=exp_url,
            errcode=status_code,
            errmsg=reason_phrase,
            headers=headers,
        )
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            async for chunk in transport._request(mock_data):
                pass
            assert exp_chunks != []

    else:
        async for chunk in transport._request(mock_data):
            assert chunk is exp_chunks.pop(0)
        assert exp_chunks == []

    assert transport._http_client.stream.call_args_list == [call('POST', transport._url, content=mock_data)]


@pytest.mark.asyncio
async def test_ScgiTransportBase_close(mocker):
    class ScgiTestTransport(_rtorrent._ScgiTransportBase):
        _get_reader_writer = AsyncMock()

    transport = ScgiTestTransport()
    await transport.close()

@pytest.mark.asyncio
async def test_ScgiTransportBase_request(mocker):
    mock_reader, mock_writer = Mock(), Mock()

    class ScgiTestTransport(_rtorrent._ScgiTransportBase):
        _get_reader_writer = AsyncMock(return_value=(mock_reader, mock_writer))

    chunks = ('foo', 'bar', 'baz')
    transport = ScgiTestTransport()
    mocker.patch.object(transport, '_send', AsyncMock())
    mocker.patch.object(transport, '_read', Mock(return_value=AsyncIterator(chunks)))

    mocks = Mock(
        _get_reader_writer=transport._get_reader_writer,
        _send=transport._send,
        _read=transport._read,
    )

    mock_data = 'mock data'
    exp_chunks = list(chunks)
    async for cunk in transport.request(mock_data):
        assert cunk is exp_chunks.pop(0)
    assert exp_chunks == []

    assert mocks.mock_calls == [
        call._get_reader_writer(),
        call._send(mock_writer, mock_data),
        call._read(mock_reader, mock_writer, 1024),
    ]

@pytest.mark.parametrize('payload_length', range(1, 16))
@pytest.mark.parametrize('headers_length', range(1, 16))
@pytest.mark.parametrize('chunk_size', (8,))
@pytest.mark.asyncio
async def test_ScgiTransportBase_read(chunk_size, headers_length, payload_length, mocker):
    headers = (b'Status: 200 OK\r\n' * headers_length)[:headers_length]
    print('headers:', headers)
    assert len(headers) == headers_length

    payload = (b'mock payload ' * payload_length)[:payload_length]
    print('payload:', payload)
    assert len(payload) == payload_length

    headers_delim = b'\r\n\r\n'
    stream = headers + headers_delim + payload
    print('stream:', stream)
    stream_chunks = [stream[i:i + chunk_size]
                     for i in range(0, len(stream), chunk_size)]
    print('stream chunks:', stream_chunks)

    first_payload_chunk_size = chunk_size - ((len(headers) + len(headers_delim)) % chunk_size)
    print('first_payload_chunk_size:', first_payload_chunk_size)
    first_payload_chunk = payload[:first_payload_chunk_size]
    print('first_payload_chunk:', first_payload_chunk)
    remaining_payload_chunks = [
        payload[i:i + chunk_size]
        for i in range(len(first_payload_chunk), len(payload), chunk_size)
    ]
    print('remaining_payload_chunks:', remaining_payload_chunks)
    exp_payload_chunks = [first_payload_chunk] + remaining_payload_chunks
    print('exp_payload_chunks:', exp_payload_chunks)

    def get_next_chunk(given_chunk_size, _my_stream_chunks=list(stream_chunks)):
        assert given_chunk_size == chunk_size
        if _my_stream_chunks:
            return _my_stream_chunks.pop(0)
        else:
            return b''

    mocks = Mock(
        read=AsyncMock(side_effect=get_next_chunk),
        close=Mock(),
        wait_closed=AsyncMock(),
    )
    mock_reader = Mock(read=mocks.read)
    mock_writer = Mock(close=mocks.close, wait_closed=mocks.wait_closed)

    class ScgiTestTransport(_rtorrent._ScgiTransportBase):
        _get_reader_writer = AsyncMock(return_value=(mock_reader, mock_writer))

    transport = ScgiTestTransport()
    async for chunk in transport._read(mock_reader, mock_writer, chunk_size):
        print('read chunk:', chunk)
        assert chunk == exp_payload_chunks.pop(0)
    assert exp_payload_chunks == []

    assert mocks.mock_calls == (
        [call.read(chunk_size)] * (len(stream_chunks) + 1)
        + [call.close(), call.wait_closed()]
    )


@pytest.mark.asyncio
async def test_ScgiTransportBase_send(mocker):
    class ScgiTestTransport(_rtorrent._ScgiTransportBase):
        _get_reader_writer = AsyncMock()

    transport = ScgiTestTransport()

    mocks = Mock(
        _encode_request=Mock(),
        write=Mock(),
        drain=AsyncMock(),
    )
    mocker.patch.object(transport, '_encode_request', mocks._encode_request)
    writer = Mock(
        write=mocks.write,
        drain=mocks.drain,
    )
    data = 'mock data'

    await transport._send(writer, data)

    assert mocks.mock_calls == [
        call._encode_request(data),
        call.write(mocks._encode_request.return_value),
        call.drain(),
    ]


@pytest.mark.parametrize(
    argnames='data, path',
    argvalues=(
        (b'foo', b'/RPC2'),
        (b'bar', b'/RPC3'),
    ),
)
def test_ScgiTransportBase_encode_request(data, path):
    class ScgiTestTransport(_rtorrent._ScgiTransportBase):
        _get_reader_writer = AsyncMock()
        _path = path

    exp_headers = (
        b'CONTENT_LENGTH\x003\x00'
        b'SCGI\x001\x00'
        b'REQUEST_METHOD\x00POST\x00'
        b'REQUEST_URI\x00' + path + b'\x00'
    )
    exp_request = (str(len(exp_headers)).encode() + b':' + exp_headers + b',' + data)

    transport = ScgiTestTransport()
    request = transport._encode_request(data)
    assert request == exp_request


@pytest.mark.parametrize(
    argnames='url, exp_host, exp_port, exp_path, proxy_url, exp_exception',
    argvalues=(
        ('http://foo:123', None, None, None, None, _errors.ValueError('Unsupported protocol: http')),
        ('scgi://foo', None, None, None, None, _errors.ValueError('No port specified')),
        ('scgi://foo:123', 'foo', 123, b'/RPC2', None, None),
        ('scgi://foo:123/bar', 'foo', 123, b'/bar', None, None),
        ('scgi://foo:123', 'foo', 123, b'/RPC2', 'http://proxy', None),
    ),
)
def test_ScgiHostTransport(url, exp_host, exp_port, exp_path, proxy_url, exp_exception):
    url = _utils.URL(url)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            _rtorrent._ScgiHostTransport(url, proxy_url)
    else:
        transport = _rtorrent._ScgiHostTransport(url, proxy_url)
        assert isinstance(transport, _rtorrent._ScgiTransportBase)
        assert transport._host == exp_host
        assert transport._port == exp_port
        assert transport._path == exp_path
        assert transport._proxy_url is proxy_url


@pytest.mark.parametrize(
    argnames='proxy_url, host, port, exception, exp_exception',
    argvalues=(
        ('http://proxy', 'localhost', 123, ValueError('Bad URL'), _errors.ValueError('Bad URL')),
        ('http://proxy', 'localhost', 123, None, None),
        ('http://a:b@proxy', 'localhost', 123, None, None),
        (None, 'localhost', 123, None, None),
    ),
)
@pytest.mark.asyncio
async def test_ScgiHostTransport_get_reader_writer(proxy_url, host, port, exception, exp_exception, mocker):
    transport = _rtorrent._ScgiHostTransport(
        url=_utils.URL(f'scgi://{host}:{port}'),
        proxy_url=_utils.URL(proxy_url) if proxy_url else None,
    )

    if exception:
        Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url',
                                           side_effect=exception)
    else:
        Proxy_from_url_mock = mocker.patch('python_socks.async_.asyncio.Proxy.from_url', return_value=Mock(
            connect=AsyncMock(return_value='mock sock'),
        ))
    reader_mock, writer_mock = Mock(), Mock()
    open_connection_mock = mocker.patch('asyncio.open_connection', AsyncMock(
        return_value=(reader_mock, writer_mock),
    ))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await transport._get_reader_writer()
    else:
        reader, writer = await transport._get_reader_writer()
        assert reader is reader_mock
        assert writer is writer_mock
        if proxy_url:
            assert Proxy_from_url_mock.call_args_list == [call(proxy_url)]
            assert Proxy_from_url_mock.return_value.connect.call_args_list == [call(
                dest_host=host,
                dest_port=port,
                timeout=float('inf'),
            )]
            assert open_connection_mock.call_args_list == [call(
                sock='mock sock',
            )]
        else:
            assert open_connection_mock.call_args_list == [call(
                host=host,
                port=port,
            )]


@pytest.mark.parametrize(
    argnames='url, exp_socket_path, exp_path, exp_exception',
    argvalues=(
        ('http://path/to/socket', None, None, _errors.ValueError('Unsupported protocol: http')),
        ('file://relative/path/to/socket', 'relative/path/to/socket', b'/RPC2', None),
        ('file:///absolute/path/to/socket', '/absolute/path/to/socket', b'/RPC2', None),
    ),
)
def test_ScgiSocketTransport(url, exp_socket_path, exp_path, exp_exception):
    print(url, exp_socket_path, exp_path, exp_exception)

    url = _utils.URL(url)
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            _rtorrent._ScgiSocketTransport(url)
    else:
        transport = _rtorrent._ScgiSocketTransport(url)
        assert isinstance(transport, _rtorrent._ScgiTransportBase)
        assert transport._socket_path == exp_socket_path
        assert transport._path == exp_path


@pytest.mark.asyncio
async def test_ScgiSocketTransport_get_reader_writer(mocker):
    transport = _rtorrent._ScgiSocketTransport(_utils.URL('file:///path/to/socket'))
    reader_mock, writer_mock = Mock(), Mock()
    open_unix_connection_mock = mocker.patch('asyncio.open_unix_connection', AsyncMock(
        return_value=(reader_mock, writer_mock),
    ))

    reader, writer = await transport._get_reader_writer()

    assert reader is reader_mock
    assert writer is writer_mock
    assert open_unix_connection_mock.call_args_list == [call(path='/path/to/socket')]



import re
from unittest.mock import Mock, call

import pytest

from aiobtclientrpc import RPCBase, _errors, _transmission, _utils

from .common import AsyncMock, make_url_parts


@pytest.mark.parametrize(
    argnames='url, exp',
    argvalues=(
        ('',
         {'scheme': 'http', 'host': 'localhost', 'port': '9091', 'path': '/transmission/rpc', 'username': None, 'password': None}),
        ('myhost',
         {'scheme': 'http', 'host': 'myhost', 'port': '9091', 'path': '/transmission/rpc', 'username': None, 'password': None}),
        ('myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': '/transmission/rpc', 'username': None, 'password': None}),
        ('myhost:123/my/path',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': '/my/path', 'username': None, 'password': None}),
        ('myhost/my/path',
         {'scheme': 'http', 'host': 'myhost', 'port': '9091', 'path': '/my/path', 'username': None, 'password': None}),
        ('foo:bar@myhost',
         {'scheme': 'http', 'host': 'myhost', 'port': '9091', 'path': '/transmission/rpc', 'username': 'foo', 'password': 'bar'}),
        ('foo:bar@myhost:123',
         {'scheme': 'http', 'host': 'myhost', 'port': '123', 'path': '/transmission/rpc', 'username': 'foo', 'password': 'bar'}),
        ('https://myhost',
         {'scheme': 'https', 'host': 'myhost', 'port': '9091', 'path': '/transmission/rpc', 'username': None, 'password': None}),
        ('file://myhost',
         _errors.ValueError('Scheme must be "http" or "https"')),
    ),
    ids=lambda v: str(v),
)
def test_TransmissionURL(url, exp):
    if isinstance(exp, Exception):
        with pytest.raises(type(exp), match=rf'^{re.escape(str(exp))}$'):
            _transmission.TransmissionURL(url)
    else:
        url = _transmission.TransmissionURL(url)
        assert make_url_parts(url) == exp


@pytest.mark.parametrize('url', (None, 'http://a:b@foo:123/custom/path'))
@pytest.mark.parametrize(
    argnames='kwargs',
    argvalues=(
        {},
        {'scheme': 'https'},
        {'host': 'asdf'},
        {'port': '123'},
        {'path': 'some/path'},
        {'username': 'this', 'password': 'that'},
        {'timeout': 123},
        {'proxy_url': 'http://hey:ho@bar:456'},
    ),
    ids=lambda v: str(v),
)
def test_instantiation(kwargs, url):
    if url:
        kwargs['url'] = url
    rpc = _transmission.TransmissionRPC(**kwargs)

    default_url = _utils.URL(_transmission.TransmissionURL.default)
    exp_url = _utils.URL(_transmission.TransmissionURL.default)
    if url:
        custom_url = _utils.URL(url)
    for name in ('scheme', 'host', 'port', 'path', 'username', 'password'):
        if name in kwargs:
            exp_value = kwargs[name]
        elif url:
            exp_value = getattr(custom_url, name)
        else:
            exp_value = getattr(default_url, name)
        actual_value = getattr(rpc.url, name)
        assert actual_value == exp_value
        setattr(exp_url, name, exp_value)

    assert rpc.url == exp_url
    assert rpc.timeout == kwargs.get('timeout', RPCBase.default_timeout)
    assert rpc.proxy_url == (_utils.URL(kwargs['proxy_url']) if 'proxy_url' in kwargs else None)

@pytest.mark.parametrize(
    argnames='kwargs, exp_error',
    argvalues=(
        ({'url': 'bar:baz'}, 'Invalid port'),
        ({'port': (1, 2, 3)}, 'Invalid port'),
        ({'timeout': 'never'}, 'Invalid timeout'),
        ({'proxy_url': 'foo://bar:baz'}, 'Invalid port'),
    ),
    ids=lambda v: str(v),
)
def test_instantiation_with_invalid_argument(kwargs, exp_error):
    with pytest.raises(_errors.ValueError, match=rf'^{re.escape(exp_error)}$'):
        _transmission.TransmissionRPC(**kwargs)


@pytest.mark.parametrize(
    argnames='method, tag, parameters, exp_json, exp_exception',
    argvalues=(
        ('foo', None, {}, '{"method": "foo"}', None),
        ('foo', 123.4, {}, '{"method": "foo", "tag": 123}', None),
        ('foo', '123', {}, '{"method": "foo", "tag": 123}', None),
        ('foo', '123.4', {}, '{"method": "foo", "tag": 123}', None),
        ('foo', 123, {'this': 'that'}, '{"method": "foo", "arguments": {"this": "that"}, "tag": 123}', None),
        ('foo', None, {'this': 123}, '{"method": "foo", "arguments": {"this": 123}}', None),
        ('foo', None, {'this': (1, 2, 'three')}, '{"method": "foo", "arguments": {"this": [1, 2, "three"]}}', None),
        ('foo', 'hey', {}, None, _errors.ValueError("Tag must be a number: 'hey'")),
        ('foo', None, {'asdf': Exception()}, None,
         _errors.ValueError("Failed to serialize to JSON: "
                            "{'method': 'foo', 'arguments': {'asdf': Exception()}}")),
    ),
)
@pytest.mark.asyncio
async def test_request_sends_and_receives_expected_data(method, tag, parameters, exp_json, exp_exception, mocker):
    rpc = _transmission.TransmissionRPC()
    mocker.patch.object(rpc, '_send_post_request', AsyncMock(return_value=Mock(status_code=200)))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._request(method, tag=tag, **parameters)
        assert rpc._send_post_request.call_args_list == []

    else:
        return_value = await rpc._request(method, tag=tag, **parameters)
        assert return_value is rpc._send_post_request.return_value
        assert rpc._send_post_request.call_args_list == [call(str(rpc.url), data=exp_json)]


@pytest.mark.parametrize(
    argnames='responses, exp_send_request_call_count, exp_exception',
    argvalues=(
        (
            [
                Mock(status_code=_transmission.TransmissionRPC._csrf_error_code,
                     headers={_transmission.TransmissionRPC._csrf_header: 'd34db33f'}),
                Mock(status_code=200),
            ],
            2,
            None
        ),
        (
            [
                Mock(status_code=_transmission.TransmissionRPC._csrf_error_code,
                     headers={_transmission.TransmissionRPC._csrf_header: 'd34db33f'}),
                Mock(status_code=_transmission.TransmissionRPC._csrf_error_code,
                     headers={_transmission.TransmissionRPC._csrf_header: 'd34db33f'},
                     __repr__=Mock(return_value='this should be HTTP status code 200')),
                Mock(status_code=200),
            ],
            2,
            RuntimeError('Unexpected response: this should be HTTP status code 200'),
        ),
        (
            [
                Mock(status_code=_transmission.TransmissionRPC._auth_error_code),
                Mock(status_code=200),
            ],
            1,
            _errors.AuthenticationError('Authentication failed'),
        ),
        (
            [
                Mock(status_code=123, __repr__=Mock(return_value='unexpected HTTP status code')),
            ],
            1,
            RuntimeError('Unexpected response: unexpected HTTP status code'),
        ),
    ),
    ids=lambda v: repr(v),
)
@pytest.mark.asyncio
async def test_request_handles_HTTP_status_codes(responses, exp_send_request_call_count, exp_exception, mocker):
    rpc = _transmission.TransmissionRPC()
    mocker.patch.object(rpc, '_send_post_request', AsyncMock(side_effect=responses))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._request('foo')
    else:
        return_value = await rpc._request('foo')
        assert return_value is responses[-1]

    assert rpc._send_post_request.call_args_list == [
        call(str(rpc.url), data='{"method": "foo"}'),
    ] * exp_send_request_call_count


@pytest.mark.asyncio
async def test_connect(mocker):
    rpc = _transmission.TransmissionRPC()
    mocker.patch.object(rpc, '_request', AsyncMock())
    await rpc._connect()
    assert rpc._request.call_args_list == [call('session-stats')]


@pytest.mark.asyncio
async def test_disconnect():
    rpc = _transmission.TransmissionRPC()
    rpc._http_headers.update({'foo': 'bar', 'baz': 123})
    assert rpc._http_headers == {'foo': 'bar', 'baz': 123}
    await rpc._disconnect()
    assert rpc._http_headers == {}


@pytest.mark.parametrize(
    argnames='method, args, kwargs, exp_params, response, exp_exception',
    argvalues=(
        (
            'some_method',
            None,
            {'foo': 'bar'},
            {'foo': 'bar'},
            Mock(json=Mock(side_effect=ValueError()), text='The Error.'),
            _errors.RPCError('Unexpected response: The Error.'),
        ),
        (
            'some_method',
            None,
            {'foo': 'bar'},
            {'foo': 'bar'},
            Mock(json=Mock(return_value={'result': 'no success'})),
            None,
        ),
        (
            'some_method',
            None,
            {'foo': 'bar'},
            {'foo': 'bar'},
            Mock(json=Mock(return_value={'result': 'success'})),
            None,
        ),
        (
            'some_method',
            {'impossible-keyword': 'important value', 'hey': 'HO!'},
            {'foo': 'bar', 'hey': 'ho'},
            {'foo': 'bar', 'hey': 'HO!', 'impossible-keyword': 'important value'},
            Mock(json=Mock(return_value={'result': 'success'})),
            None,
        ),
    ),
    ids=lambda v: str(v),
)
@pytest.mark.asyncio
async def test_call(method, args, kwargs, exp_params, response, exp_exception, mocker):
    rpc = _transmission.TransmissionRPC()
    mocker.patch.object(rpc, '_request', AsyncMock(return_value=response))

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await rpc._call(method, args=args, **kwargs)
    else:
        return_value = await rpc._call(method, args=args, **kwargs)
        assert return_value is response.json.return_value

    assert rpc._request.call_args_list == [call(method, **exp_params)]
